"""Genshin transaction models."""

import enum
import typing

from genshin.models.model import Aliased, APIModel, TZDateTime, Unique

__all__ = ["BaseTransaction", "ItemTransaction", "Transaction", "TransactionKind"]


class TransactionKind(str, enum.Enum):
    """Possible kind of transaction."""

    PRIMOGEM = "primogem"
    """Primogem currency."""

    CRYSTAL = "crystal"
    """Genesis crystal currency."""

    RESIN = "resin"
    """Resin currency."""

    ARTIFACT = "artifact"
    """Artifact items from domains."""

    WEAPON = "weapon"
    """Weapon items from domains and wishes."""


class BaseTransaction(APIModel, Unique):
    """Genshin transaction."""

    kind: TransactionKind

    id: int
    time: TZDateTime = Aliased("datetime")
    amount: int = Aliased("add_num")
    reason: str = Aliased("reason")


class Transaction(BaseTransaction):
    """Genshin transaction of currency."""

    kind: typing.Literal[TransactionKind.PRIMOGEM, TransactionKind.CRYSTAL, TransactionKind.RESIN]


class ItemTransaction(BaseTransaction):
    """Genshin transaction of artifacts or weapons."""

    kind: typing.Literal[TransactionKind.ARTIFACT, TransactionKind.WEAPON]

    name: str
    rarity: int = Aliased("quality")



"""StarRail battle chronicle component."""

import asyncio
import functools
import typing

from genshin import errors, paginators, types, utility
from genshin.client import routes
from genshin.models import zzz as models
from genshin.models.genshin import gacha as gacha_models

from . import base

__all__ = ("ZZZBattleChronicleClient",)


class ZZZBattleChronicleClient(base.BaseBattleChronicleClient):
    """ZZZ battle chronicle component."""

    async def _request_zzz_record(
        self,
        endpoint: str,
        uid: typing.Optional[int] = None,
        *,
        method: str = "GET",
        lang: typing.Optional[str] = None,
        payload: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        cache: bool = False,
        is_nap_ledger: bool = False,
        use_uid_in_payload: bool = False,
    ) -> typing.Mapping[str, typing.Any]:
        """Get an arbitrary ZZZ object."""
        payload = dict(payload or {})
        original_payload = payload.copy()

        uid = uid or await self._get_uid(types.Game.ZZZ)

        if is_nap_ledger or use_uid_in_payload:
            payload = {
                "uid": uid,
                "region": utility.recognize_zzz_server(uid),
                **payload,
            }
        else:
            payload = {
                "role_id": uid,
                "server": utility.recognize_zzz_server(uid),
                **payload,
            }

        data, params = None, None
        if method == "POST":
            data = payload
        else:
            params = payload

        cache_key: typing.Optional[base.ChronicleCacheKey] = None
        if cache:
            cache_key = base.ChronicleCacheKey(
                types.Game.ZZZ,
                endpoint,
                uid,
                lang=lang or self.lang,
                params=tuple(original_payload.values()),
            )

        return await self.request_game_record(
            endpoint,
            lang=lang,
            game=types.Game.ZZZ,
            region=utility.recognize_region(uid, game=types.Game.ZZZ),
            params=params,
            data=data,
            cache=cache_key,
            custom_route=routes.NAP_LEDGER_URL if is_nap_ledger else None,
        )

    @typing.overload
    async def get_zzz_notes(
        self,
        uid: typing.Optional[int] = ...,
        *,
        lang: typing.Optional[str] = ...,
        autoauth: bool = ...,
        return_raw_data: typing.Literal[False] = ...,
    ) -> models.ZZZNotes: ...
    @typing.overload
    async def get_zzz_notes(
        self,
        uid: typing.Optional[int] = ...,
        *,
        lang: typing.Optional[str] = ...,
        autoauth: bool = ...,
        return_raw_data: typing.Literal[True] = ...,
    ) -> typing.Mapping[str, typing.Any]: ...
    async def get_zzz_notes(
        self,
        uid: typing.Optional[int] = None,
        *,
        lang: typing.Optional[str] = None,
        autoauth: bool = True,
        return_raw_data: bool = False,
    ) -> typing.Union[models.ZZZNotes, typing.Mapping[str, typing.Any]]:
        """Get ZZZ sticky notes (real-time notes)."""
        try:
            data = await self._request_zzz_record("note", uid, lang=lang)
        except errors.DataNotPublic as e:
            # error raised only when real-time notes are not enabled
            if uid and (await self._get_uid(types.Game.ZZZ)) != uid:
                raise errors.GenshinException(e.response, "Cannot view real-time notes of other users.") from e
            if not autoauth:
                raise errors.GenshinException(e.response, "Real-time notes are not enabled.") from e

            await self.update_settings(3, True, game=types.Game.ZZZ)
            data = await self._request_zzz_record("note", uid, lang=lang)

        if return_raw_data:
            return data
        return models.ZZZNotes(**data)

    async def get_zzz_diary(
        self,
        uid: typing.Optional[int] = None,
        *,
        month: typing.Optional[str] = None,
        lang: typing.Optional[str] = None,
    ) -> models.ZZZDiary:
        """Get ZZZ inter-knot monthly earning data."""
        data = await self._request_zzz_record(
            "month_info", uid, lang=lang, payload={"month": month or ""}, is_nap_ledger=True
        )
        return models.ZZZDiary(**data)

    async def get_zzz_diary_detail(
        self,
        month: str,
        *,
        type: models.ZZZCurrencyType,
        page: int = 1,
        page_size: int = 20,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> models.ZZZDiaryDetail:
        """Get ZZZ inter-knot monthly earning data."""
        if not month:
            raise ValueError("month is required.")

        data = await self._request_zzz_record(
            "month_detail",
            uid,
            lang=lang,
            payload={"month": month, "current_page": page, "type": type.value, "page_size": page_size},
            is_nap_ledger=True,
        )
        return models.ZZZDiaryDetail(**data)

    async def get_zzz_user(
        self,
        uid: typing.Optional[int] = None,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.ZZZUserStats:
        """Get ZZZ user stats."""
        data = await self._request_zzz_record("index", uid, lang=lang)
        return models.ZZZUserStats(**data)

    async def get_zzz_agents(
        self, uid: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> typing.Sequence[models.ZZZPartialAgent]:
        """Get all owned ZZZ characters (only brief info)."""
        data = await self._request_zzz_record("avatar/basic", uid, lang=lang)
        return [models.ZZZPartialAgent(**item) for item in data["avatar_list"]]

    async def get_bangboos(
        self, uid: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> typing.Sequence[models.ZZZBaseBangboo]:
        """Get all owned ZZZ bangboos."""
        data = await self._request_zzz_record("buddy/info", uid, lang=lang)
        return [models.ZZZBaseBangboo(**item) for item in data["list"]]

    @typing.overload
    async def get_zzz_agent_info(
        self,
        character_id: int,
        *,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> models.ZZZFullAgent: ...
    @typing.overload
    async def get_zzz_agent_info(
        self,
        character_id: typing.Sequence[int],
        *,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.ZZZFullAgent]: ...
    async def get_zzz_agent_info(
        self,
        character_id: typing.Union[int, typing.Sequence[int]],
        *,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Union[models.ZZZFullAgent, typing.Sequence[models.ZZZFullAgent]]:
        """Get a ZZZ character's detailed info."""
        if isinstance(character_id, typing.Sequence):
            tasks = [
                self._request_zzz_record("avatar/info", uid, lang=lang, payload={"id_list[]": character_id_})
                for character_id_ in character_id
            ]
            results = await asyncio.gather(*tasks)
            return [models.ZZZFullAgent(**data["avatar_list"][0]) for data in results]

        data = await self._request_zzz_record("avatar/info", uid, lang=lang, payload={"id_list[]": character_id})
        return models.ZZZFullAgent(**data["avatar_list"][0])

    @typing.overload
    async def get_shiyu_defense(
        self,
        uid: typing.Optional[int] = ...,
        *,
        previous: bool = ...,
        lang: typing.Optional[str] = ...,
        raw: typing.Literal[False] = ...,
    ) -> typing.Union[models.ShiyuDefenseV1, models.ShiyuDefenseV2]: ...
    @typing.overload
    async def get_shiyu_defense(
        self,
        uid: typing.Optional[int] = ...,
        *,
        previous: bool = ...,
        lang: typing.Optional[str] = ...,
        raw: typing.Literal[True] = ...,
    ) -> typing.Mapping[str, typing.Any]: ...
    async def get_shiyu_defense(
        self,
        uid: typing.Optional[int] = None,
        *,
        previous: bool = False,
        lang: typing.Optional[str] = None,
        raw: bool = False,
    ) -> typing.Union[models.ShiyuDefenseV1, models.ShiyuDefenseV2, typing.Mapping[str, typing.Any]]:
        """Get ZZZ Shiyu defense stats."""
        payload = {"schedule_type": 2 if previous else 1}
        data = await self._request_zzz_record("hadal_info_v2", uid, lang=lang, payload=payload)
        version = data["hadal_ver"]  # v1 or v2
        key = f"hadal_info_{version}"

        if version == "v1":
            data = data[key]
        elif version == "v2":
            nickname = data["nick_name"]
            icon = data["icon"]
            data = data[key]
            data["nick_name"] = nickname
            data["icon"] = icon
        else:
            msg = f"Unknown Shiyu Defense version: {version!r}"
            raise ValueError(msg)

        self._add_timezone_to_data(data, ("hadal_begin_time", "hadal_end_time"), game=types.Game.ZZZ, uid=uid)

        if raw:
            return data
        if version == "v2":
            return models.ShiyuDefenseV2(**data)
        return models.ShiyuDefenseV1(**data)

    @typing.overload
    async def get_deadly_assault(
        self,
        uid: typing.Optional[int] = ...,
        *,
        previous: bool = ...,
        lang: typing.Optional[str] = ...,
        raw: typing.Literal[False] = ...,
    ) -> models.DeadlyAssault: ...
    @typing.overload
    async def get_deadly_assault(
        self,
        uid: typing.Optional[int] = ...,
        *,
        previous: bool = ...,
        lang: typing.Optional[str] = ...,
        raw: typing.Literal[True] = ...,
    ) -> typing.Mapping[str, typing.Any]: ...
    async def get_deadly_assault(
        self,
        uid: typing.Optional[int] = None,
        *,
        previous: bool = False,
        lang: typing.Optional[str] = None,
        raw: bool = False,
    ) -> typing.Union[models.DeadlyAssault, typing.Mapping[str, typing.Any]]:
        """Get ZZZ Shiyu defense stats."""
        payload = {"schedule_type": 2 if previous else 1}
        data = await self._request_zzz_record("mem_detail", uid, lang=lang, payload=payload, use_uid_in_payload=True)

        self._add_timezone_to_data(data, ("hadal_begin_time", "hadal_end_time"), game=types.Game.ZZZ, uid=uid)

        if raw:
            return data
        return models.DeadlyAssault(**data)

    async def get_lost_void_summary(
        self, uid: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> models.LostVoidSummary:
        """Get ZZZ Lost Void summary."""
        data = await self._request_zzz_record("abysss2_abstract", uid, lang=lang, use_uid_in_payload=True)
        return models.LostVoidSummary(**data)

    async def get_threshold_simulation_brief(
        self, uid: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> models.ThresholdSimulationInfo:
        """Get ZZZ Threshold Simulation brief info."""
        data = await self._request_zzz_record(
            "void_front_battle_abstract_info", uid, lang=lang, use_uid_in_payload=True
        )
        return models.ThresholdSimulationInfo(**data["void_front_battle_abstract_info_brief"])

    @typing.overload
    async def get_threshold_simulation(
        self,
        id: typing.Optional[int] = ...,
        uid: typing.Optional[int] = ...,
        *,
        lang: typing.Optional[str] = ...,
        raw: typing.Literal[False] = ...,
    ) -> models.ThresholdSimulation: ...
    @typing.overload
    async def get_threshold_simulation(
        self,
        id: typing.Optional[int] = ...,
        uid: typing.Optional[int] = ...,
        *,
        lang: typing.Optional[str] = ...,
        raw: typing.Literal[True] = ...,
    ) -> typing.Mapping[str, typing.Any]: ...
    async def get_threshold_simulation(
        self,
        id: typing.Optional[int] = None,
        uid: typing.Optional[int] = None,
        *,
        lang: typing.Optional[str] = None,
        raw: bool = False,
    ) -> typing.Union[models.ThresholdSimulation, typing.Mapping[str, typing.Any]]:
        """Get ZZZ Threshold Simulation stats.

        If no ID is given, the latest run will be fetched.
        """
        if id is None:
            brief = await self.get_threshold_simulation_brief(uid, lang=lang)
            id = brief.id

        data = await self._request_zzz_record(
            "void_front_battle_detail", uid, lang=lang, payload={"void_front_id": id}, use_uid_in_payload=True
        )
        if raw:
            return data
        return models.ThresholdSimulation(**data)

    async def _get_chronicle_signal_page(
        self,
        end_id: int,
        banner_type: gacha_models.ZZZBannerType,
        *,
        lang: typing.Optional[str] = None,
        uid: typing.Optional[int] = None,
    ) -> typing.Sequence[gacha_models.SignalSearch]:
        """Get a single page of battle chronicle signal searches."""
        uid = uid or await self._get_uid(types.Game.ZZZ)
        timezone = self.get_account_timezone(game=types.Game.ZZZ, uid=uid)
        if timezone is None:
            msg = "Cannot find account timezone."
            raise ValueError(msg)

        data = await self._request_zzz_record(
            "gacha_record",
            uid,
            lang=lang,
            payload={"gacha_type": banner_type.to_chronicle_type(), "end_id": end_id},
            use_uid_in_payload=True,
        )
        records = data.get("gacha_item_list", [])
        return [
            gacha_models.SignalSearch.from_chronicle_data(i, uid=uid, banner_type=banner_type, tz_offset=timezone - 8)
            for i in records
        ]

    def chronicle_signal_history(
        self,
        banner_type: typing.Optional[typing.Union[int, typing.Sequence[int]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        uid: typing.Optional[int] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[gacha_models.SignalSearch]:
        """Get the signal search history of a user."""
        banner_types = banner_type or list(gacha_models.ZZZBannerType)

        if not isinstance(banner_types, typing.Sequence):
            banner_types = [banner_types]

        iterators: list[paginators.Paginator[gacha_models.SignalSearch]] = []
        for banner in banner_types:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_chronicle_signal_page,
                        banner_type=gacha_models.ZZZBannerType(banner),
                        lang=lang,
                        uid=uid,
                    ),
                    limit=limit,
                    end_id=end_id,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda wish: wish.time.timestamp())

    async def get_zzz_event_calendar(
        self, uid: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> typing.Sequence[models.ZZZEvent]:
        """Get ZZZ event calendar."""
        data = await self._request_zzz_record("activity_calendar", uid, lang=lang, use_uid_in_payload=True)
        return [models.ZZZEvent(**item) for item in data["activity_list"]]

    async def get_zzz_gacha_calendar(
        self, uid: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> models.ZZZGachaCalendar:
        """Get ZZZ gacha calendar."""
        data = await self._request_zzz_record("gacha_calendar", uid, lang=lang, use_uid_in_payload=True)
        return models.ZZZGachaCalendar(**data)

    async def get_zzz_gacha_info(
        self, uid: typing.Optional[int] = None, *, lang: typing.Optional[str] = None
    ) -> models.ZZZGachaInfo:
        """Get ZZZ gacha info."""
        data = await self._request_zzz_record("cur_gacha_detail", uid, lang=lang, use_uid_in_payload=True)
        return models.ZZZGachaInfo(**data)



"""Chronicle activities models."""

import datetime
import re
import typing

import pydantic

from genshin.models.genshin import character
from genshin.models.model import Aliased, APIModel

__all__ = [
    "Activities",
    "Activity",
    "EnergyAmplifier",
    "HyakuninIkki",
    "LabyrinthWarriors",
    "OldActivity",
    "Potion",
    "Summer",
]

ModelT = typing.TypeVar("ModelT", bound=APIModel)


class OldActivity(APIModel, typing.Generic[ModelT]):
    """Arbitrary activity for chinese events."""

    # sometimes __parameters__ may not be provided in older versions
    __parameters__: typing.ClassVar[tuple[typing.Any, ...]] = (ModelT,)  # type: ignore

    exists_data: bool
    records: typing.Sequence[ModelT]


class Activity(OldActivity[ModelT], typing.Generic[ModelT]):
    """Arbitrary activity for chinese events."""

    start_time: typing.Optional[datetime.datetime] = None
    end_time: typing.Optional[datetime.datetime] = None
    total_score: int = 0

    # cn only
    total_times: int = 1


# ---------------------------------------------------------
# Hyakunin Ikki:


class HyakuninIkkiCharacter(character.BaseCharacter):
    """Possibly trial Hyakunin Ikki character."""

    level: int
    trial: bool = Aliased("is_trail_avatar")


class HyakuninIkkiSkill(APIModel):
    """Hyakunin Ikki skill."""

    id: int
    name: str
    icon: str
    description: str = Aliased("desc")


class HyakuninIkkiBattle(APIModel):
    """Hyakunin Ikki battle."""

    characters: typing.Sequence[HyakuninIkkiCharacter] = Aliased("avatars")
    skills: typing.Sequence[HyakuninIkkiSkill] = Aliased("skills")

    @pydantic.field_validator("characters", mode="before")
    def __validate_characters(cls, value: typing.Sequence[typing.Any]) -> typing.Sequence[typing.Any]:
        """Remove characters with a null id."""
        return [character for character in value if character["id"]]


class HyakuninIkki(APIModel):
    """Hyakunin Ikki challenge."""

    id: int = Aliased("challenge_id")
    name: str = Aliased("challenge_name")
    difficulty: int
    multiplier: int = Aliased("score_multiple")
    score: int = Aliased("max_score")
    medal_icon: str = Aliased("heraldry_icon")

    battles: typing.Sequence[HyakuninIkkiBattle] = Aliased("lineups")

    @property
    def medal(self) -> str:
        match = re.search(r"heraldry_(\w+)\.png", self.medal_icon)
        return match.group(1) if match else ""


# ---------------------------------------------------------
# Labyrinth Warriors:


class LabyrinthWarriorsCharacter(character.BaseCharacter):
    """Labyrinth Warriors character."""

    level: int


class LabyrinthWarriorsRune(APIModel):
    """Labyrinth Warriors rune."""

    id: int
    icon: str
    name: str
    description: str = Aliased("desc")
    element: str


class LabyrinthWarriors(APIModel):
    """Labyrinth Warriors challenge."""

    id: int = Aliased("challenge_id")
    name: str = Aliased("challenge_name")
    passed: bool = Aliased("is_passed")
    level: int = Aliased("settled_level")

    main_characters: typing.Sequence[LabyrinthWarriorsCharacter] = Aliased("main_avatars")
    support_characters: typing.Sequence[LabyrinthWarriorsCharacter] = Aliased("support_avatars")
    runes: typing.Sequence[LabyrinthWarriorsRune]


# ---------------------------------------------------------
# Energy Amplifier Fruition:


class EnergyAmplifierCharacter(character.BaseCharacter):
    """Energy Amplifier character."""

    level: int


class EnergyAmplifierCriteria(APIModel):
    """Energy Amplifier criteria."""

    id: int
    description: str = Aliased("desc")
    score: int


class EnergyAmplifierBuff(APIModel):
    """Energy Amplifier buff."""

    id: int
    name: str
    quality: int
    description: str = Aliased("desc")
    energy: int


class EnergyAmplifier(APIModel):
    """Energy Amplifier challenge."""

    id: int = Aliased("challenge_id")
    name: str = Aliased("challenge_name")
    energy: int
    difficulty: int
    max_score: int
    score_multiplier: int = Aliased("score_multiple")

    characters: typing.Sequence[EnergyAmplifierCharacter] = Aliased("avatars")
    criteria: typing.Sequence[EnergyAmplifierCriteria] = Aliased("limit_conditions")
    buffs: typing.Sequence[EnergyAmplifierBuff]


# ---------------------------------------------------------
# A Study In Potions:


class PotionCharacter(APIModel):
    """Study In Potions character."""

    level: int
    trial: bool = Aliased("is_trial")


class PotionBuff(APIModel):
    """Study In Potions buff."""

    id: int
    name: str
    description: str = Aliased("desc")
    quality: int
    icon: str
    mark: str = Aliased("cornor_mark")


class PotionStage(APIModel):
    """Study In Potions stage."""

    name: str = Aliased("level_name")
    difficulty: int
    difficulty_id: int
    score: int
    score_multiplier: float = Aliased("factor")

    characters: typing.Sequence[PotionCharacter] = Aliased("avatars")
    buffs: typing.Sequence[PotionBuff]


class Potion(APIModel):
    """Study In Potions challenge."""

    name: str = Aliased("stage_name")
    score: int = Aliased("stage_score")
    finished: bool
    stages: typing.Sequence[PotionStage] = Aliased("levels")


# ---------------------------------------------------------
# Summer：


class SummerMemories(APIModel):
    """Summer story record."""

    finish_time: typing.Optional[datetime.datetime]
    finished: bool
    icon: str
    name: str

    @pydantic.field_validator("finish_time", mode="before")
    def __validate_time(cls, value: typing.Any) -> typing.Optional[datetime.datetime]:
        if value is None or isinstance(value, datetime.datetime):
            return value

        return datetime.datetime(value["year"], value["month"], value["day"], value["hour"], value["minute"])


class SummerSurfpiercer(APIModel):
    """Summer sailing record."""

    id: int
    time: int = Aliased("cost_time")
    finished: bool


class SummerRealmExploration(APIModel):
    """Summer challenge record."""

    id: int
    finish_time: typing.Optional[datetime.datetime] = Aliased("finish_time")
    finished: bool
    success: int = Aliased("success_num")
    skills_used: int = Aliased("skill_use_num")
    name: str
    icon: str

    @pydantic.field_validator("finish_time", mode="before")
    def __validate_time(cls, value: typing.Any) -> typing.Optional[datetime.datetime]:
        if value is None or isinstance(value, datetime.datetime):
            return value

        return datetime.datetime(value["year"], value["month"], value["day"], value["hour"], value["minute"])


class Summer(APIModel):
    """Summer event."""

    waverider_waypoints: int = Aliased("anchor_number")
    waypoints: int = Aliased("way_point_number")
    treasure_chests: int = Aliased("chest_number")

    surfpiercer: typing.Sequence[SummerSurfpiercer] = Aliased("sailing")
    memories: typing.Sequence[SummerMemories] = Aliased("story")
    realm_exploration: typing.Sequence[SummerRealmExploration] = Aliased("challenge")

    @pydantic.field_validator("surfpiercer", "memories", "realm_exploration", mode="before")
    def __flatten_records(cls, value: typing.Any) -> typing.Sequence[typing.Any]:
        if isinstance(value, typing.Sequence):
            return typing.cast("typing.Sequence[object]", value)

        return value["records"]


# ---------------------------------------------------------
# Activities:


class Activities(APIModel):
    """Collection of genshin activities."""

    hyakunin_ikki_v21: typing.Optional[OldActivity[HyakuninIkki]] = pydantic.Field(
        None, json_schema_extra={"gslug": "sumo"}
    )
    hyakunin_ikki_v25: typing.Optional[OldActivity[HyakuninIkki]] = pydantic.Field(
        None, json_schema_extra={"gslug": "sumo_second"}
    )
    labyrinth_warriors: typing.Optional[OldActivity[LabyrinthWarriors]] = pydantic.Field(
        None, json_schema_extra={"gslug": "rogue"}
    )
    energy_amplifier: typing.Optional[Activity[EnergyAmplifier]] = pydantic.Field(
        None, json_schema_extra={"gslug": "channeller_slab_copy"}
    )
    study_in_potions: typing.Optional[OldActivity[Potion]] = pydantic.Field(None, json_schema_extra={"gslug": "potion"})
    summertime_odyssey: typing.Optional[Summer] = pydantic.Field(None, json_schema_extra={"gslug": "summer_v2"})

    effigy: typing.Optional[Activity[typing.Any]] = None
    mechanicus: typing.Optional[Activity[typing.Any]] = None
    fleur_fair: typing.Optional[Activity[typing.Any]] = None
    channeller_slab: typing.Optional[Activity[typing.Any]] = None
    martial_legend: typing.Optional[Activity[typing.Any]] = None
    chess: typing.Optional[Activity[typing.Any]] = None

    @pydantic.model_validator(mode="before")
    def __flatten_activities(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        if not values.get("activities"):
            return values

        slugs = {  # type: ignore
            field.json_schema_extra["gslug"]: name
            for name, field in Activities.model_fields.items()
            if isinstance(field.json_schema_extra, dict) and field.json_schema_extra.get("gslug")
        }

        activites: list[dict[str, typing.Any]] = values["activities"]
        for activity in activites:
            for name, value in activity.items():
                if "exists_data" not in value:
                    continue

                name_ = slugs.get(name, name)
                values[name_] = value if value["exists_data"] else None

        return values



"""Starrail base character model."""

import enum
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, Unique

__all__ = (
    "BaseMemoSprite",
    "BaseRelic",
    "BaseSkill",
    "DetailMemoSprite",
    "DetailRelic",
    "DetailRelicProperty",
    "DetailSkill",
    "FloorCharacter",
    "LineupMemoSprite",
    "LineupRelic",
    "PropertyInfo",
    "Rank",
    "RelicProperty",
    "RogueCharacter",
    "SimpleRelic",
    "SkillStage",
    "StarRailBaseCharacter",
    "StarRailBaseEquipment",
    "StarRailBaseProperty",
    "StarRailCharacterProperty",
    "StarRailDetailCharacter",
    "StarRailEquipment",
    "StarRailLineupCharacter",
    "StarRailLineupProperty",
    "StarRailPartialCharacter",
    "StarRailPath",
    "StarRailSimpleCharacter",
)


class StarRailPath(enum.IntEnum):
    """StarRail character path."""

    DESTRUCTION = 1
    THE_HUNT = 2
    ERUDITION = 3
    HARMONY = 4
    NIHILITY = 5
    PRESERVATION = 6
    ABUNDANCE = 7
    REMEMBRANCE = 8
    ELATION = 9


class StarRailBaseCharacter(APIModel, Unique):
    """Base character model."""

    id: int
    element: str
    rarity: int
    icon: str


class StarRailPartialCharacter(StarRailBaseCharacter):
    """Character without any equipment."""

    name: str
    level: int
    rank: int


class FloorCharacter(StarRailBaseCharacter):
    """Character in a floor."""

    level: int
    rank: int


class RogueCharacter(StarRailBaseCharacter):
    """Rogue character model."""

    level: int
    rank: int


class StarRailBaseEquipment(APIModel):
    """HSR character light cone."""

    id: int
    level: int
    rank: int
    name: str
    desc: str
    icon: str
    rarity: int


class StarRailEquipment(StarRailBaseEquipment):
    """HSR character light cone."""

    wiki: str


class PropertyInfo(APIModel):
    """Relic property info."""

    property_type: int
    name: str
    icon: str
    property_name_relic: str
    property_name_filter: str


class RelicProperty(APIModel):
    """HSR relic property."""

    property_type: int
    value: str
    times: int
    name: str = Aliased("property_name")
    icon: str


class DetailRelicProperty(APIModel):
    """Relic property for detailed character."""

    property_type: int
    value: str
    times: int
    preferred: bool
    recommended: bool
    info: PropertyInfo


class BaseRelic(APIModel):
    """HSR base relic."""

    id: int
    pos: int
    rarity: int


class LineupRelic(BaseRelic):
    """HSR lineup relic."""

    set_name: str
    set_num: int

    main_property: RelicProperty
    properties: typing.Sequence[RelicProperty]

    wiki: str = Aliased("wiki_url")


class SimpleRelic(BaseRelic):
    """HSR simple relic without properties."""

    level: int
    name: str
    desc: str
    icon: str


class DetailRelic(SimpleRelic):
    """HSR character relic with properties."""

    wiki: str
    main_property: DetailRelicProperty
    properties: typing.Sequence[DetailRelicProperty]


class Rank(APIModel):
    """HSR character eidolon."""

    id: int
    pos: int
    name: str
    icon: str
    desc: str
    is_unlocked: bool


class StarRailLineupProperty(APIModel):
    """HSR character/memosprite property for lineup simulator."""

    property_type: int
    base: str
    add: str
    final: str
    name: str = Aliased("property_name")
    icon: str

    base_float: float
    add_float: float
    final_float: float


class StarRailBaseProperty(APIModel):
    """HSR detail character endpoint base property (for memosprite.)"""

    property_type: int
    info: PropertyInfo
    base: str
    add: str
    final: str


class StarRailCharacterProperty(StarRailBaseProperty):
    """HSR property for detailed character, with preferred and recommended flags."""

    preferred: bool
    recommended: bool


class SkillStage(APIModel):
    """Character skill stage."""

    name: str
    desc: str
    level: int
    remake: str
    item_url: str
    is_activated: bool
    is_rank_work: bool


class BaseSkill(APIModel):
    """HSR character/memosprite skill."""

    point_id: str
    point_type: int
    item_url: str
    level: int
    is_activated: bool
    is_rank_work: bool
    pre_point: str
    anchor: str
    remake: str


class DetailSkill(BaseSkill):
    """HSR character/memosprite skill for detail character."""

    skill_stages: typing.Sequence[SkillStage]


class BaseMemoSprite(APIModel):
    """HSR base memosprite."""

    id: int = Aliased("servant_id")
    name: str = Aliased("servant_name")
    icon: str = Aliased("servant_icon")


class LineupMemoSprite(BaseMemoSprite):
    """HSR lineup memosprite."""

    properties: typing.Sequence[StarRailLineupProperty] = Aliased("servant_properties")
    skills: typing.Sequence[BaseSkill] = Aliased("servant_skills")


class DetailMemoSprite(BaseMemoSprite):
    """HSR memosprite for detail character."""

    properties: typing.Sequence[StarRailBaseProperty] = Aliased("servant_properties")
    skills: typing.Sequence[DetailSkill] = Aliased("servant_skills")


class StarRailSimpleCharacter(StarRailPartialCharacter):
    """HSR character with simplified data.

    Relics and ornaments have no properties, no skills, properties, and memosprite.
    This happens when the character is viewed by other players.
    """

    image: str
    equip: typing.Optional[StarRailEquipment]
    relics: typing.Sequence[SimpleRelic]
    ornaments: typing.Sequence[SimpleRelic]
    ranks: typing.Sequence[Rank]
    path: StarRailPath = Aliased("base_type")
    figure_path: str

    @property
    def is_wearing_outfit(self) -> bool:
        """Whether the character is wearing an outfit."""
        return "avatar_skin_image" in self.image


class StarRailDetailCharacter(StarRailSimpleCharacter):
    """HSR character with detailed equipment and relics."""

    relics: typing.Sequence[DetailRelic]
    ornaments: typing.Sequence[DetailRelic]

    properties: typing.Sequence[StarRailCharacterProperty]
    skills: typing.Sequence[DetailSkill]
    memosprite: typing.Optional[DetailMemoSprite] = Aliased("servant_detail")

    @pydantic.field_validator("memosprite", mode="before")
    @classmethod
    def __return_none(
        cls, value: typing.Optional[typing.Dict[str, typing.Any]]
    ) -> typing.Optional[typing.Dict[str, typing.Any]]:
        """Return None if memosprite ID is 0."""
        if value is None:
            return None
        if value.get("servant_id", "0") == "0":
            return None
        return value


class StarRailLineupCharacter(StarRailPartialCharacter):
    """HSR lineup simulator character."""

    equip: typing.Optional[StarRailBaseEquipment]
    relics: typing.Sequence[LineupRelic]
    ranks: typing.Sequence[Rank]
    properties: typing.Sequence[StarRailLineupProperty]
    skills: typing.Sequence[BaseSkill]
    memosprite: typing.Optional[LineupMemoSprite] = Aliased("servant_detail")

    path: StarRailPath = Aliased("base_type")
    figure_path: str



"""Honkai chronicle battlesuits."""

import re
import typing

import pydantic

from genshin.models.honkai import battlesuit
from genshin.models.model import Aliased, APIModel, Unique


class Equipment(APIModel, Unique):
    """Battlesuit equipment."""

    id: int
    name: str
    rarity: int
    max_rarity: int
    icon: str

    @property
    def type(self) -> str:
        """The type of the equipment"""
        match = re.search(r"/(\w+)Icons/", self.icon)
        base_type = match[1] if match else ""
        if not base_type or base_type == "Stigmata":
            return base_type

        match = re.search(r"/Weapon_([A-Za-z]+?)_", self.icon)
        return match[1] if match else "Weapon"


class Stigma(Equipment):
    """Battlesuit stigma."""


class BattlesuitWeapon(Equipment):
    """Battlesuit weapon."""


class FullBattlesuit(battlesuit.Battlesuit):
    """Battlesuit complete with equipped weapon and stigmata."""

    level: int
    weapon: BattlesuitWeapon
    stigmata: typing.Sequence[Stigma] = Aliased("stigmatas")

    @pydantic.model_validator(mode="before")
    def __unnest_char_data(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        if isinstance(values.get("character"), typing.Mapping):
            values.update(values["character"])

        values.update(values.get("avatar", {}))

        return values

    @pydantic.field_validator("stigmata")
    def __remove_unequipped_stigmata(cls, value: typing.Sequence[Stigma]) -> typing.Sequence[Stigma]:
        return [stigma for stigma in value if stigma.id != 0]



"""Constants hardcoded for optimizations."""

import datetime
import typing

from . import types

__all__ = [
    "APP_IDS",
    "APP_KEYS",
    "CN_TIMEZONE",
    "DS_SALT",
    "GAME_BIZS",
    "GAME_LANGS",
    "GEETEST_RECORD_KEYS",
    "GEETEST_RETCODES",
    "LANGS",
]


LANGS: typing.Final[typing.Dict[types.Lang, str]] = {
    "zh-cn": "简体中文",
    "zh-tw": "繁體中文",
    "de-de": "Deutsch",
    "en-us": "English",
    "es-es": "Español",
    "fr-fr": "Français",
    "id-id": "Indonesia",
    "it-it": "Italiano",
    "ja-jp": "日本語",
    "ko-kr": "한국어",
    "pt-pt": "Português",
    "ru-ru": "Pусский",
    "th-th": "ภาษาไทย",
    "vi-vn": "Tiếng Việt",
    "tr-tr": "Türkçe",
}
"""Languages supported by the API."""

# https://webstatic.hoyoverse.com/admin/mi18n/plat_cn/m10201340231541/m10201340231541-zh-cn.json
GAME_LANGS: typing.Final[typing.Dict[types.Game, typing.Tuple[types.Lang, ...]]] = {
    types.Game.HONKAI: (
        "zh-cn",
        "zh-tw",
        "de-de",
        "en-us",
        "es-es",
        "fr-fr",
        "id-id",
        "ja-jp",
        "ko-kr",
        "pt-pt",
        "ru-ru",
        "th-th",
        "vi-vn",
    ),
    types.Game.GENSHIN: (
        "zh-cn",
        "zh-tw",
        "de-de",
        "en-us",
        "es-es",
        "fr-fr",
        "id-id",
        "it-it",
        "ja-jp",
        "ko-kr",
        "pt-pt",
        "ru-ru",
        "th-th",
        "tr-tr",
        "vi-vn",
    ),
    types.Game.STARRAIL: (
        "zh-cn",
        "zh-tw",
        "de-de",
        "en-us",
        "es-es",
        "fr-fr",
        "id-id",
        "ja-jp",
        "ko-kr",
        "pt-pt",
        "ru-ru",
        "th-th",
        "vi-vn",
    ),
    types.Game.ZZZ: (
        "zh-cn",
        "zh-tw",
        "de-de",
        "en-us",
        "es-es",
        "fr-fr",
        "id-id",
        "ja-jp",
        "ko-kr",
        "pt-pt",
        "ru-ru",
        "th-th",
        "vi-vn",
    ),
    types.Game.TOT: (
        "zh-cn",
        "zh-tw",
        "de-de",
        "en-us",
        "es-es",
        "fr-fr",
        "id-id",
        "ja-jp",
        "ko-kr",
        "pt-pt",
        "ru-ru",
        "th-th",
        "vi-vn",
    ),
}

DS_SALT: typing.Final[typing.Dict[typing.Union[types.Region, str], str]] = {
    types.Region.OVERSEAS: "6s25p5ox5y14umn1p61aqyyvbvvl3lrt",
    types.Region.CHINESE: "xV8v4Qu54lUKrEYFZkJhB8cuOh9Asafs",
    "app_login": "IZPgfb0dRPtBeLuFkdDznSZ6f4wWt6y2",
    "cn_signin": "LyD1rXqMv2GJhnwdvCBjFOKGiKuLY3aO",
    "cn_passport": "JwYDpKvLj6MrMqqYU6jTKF17KNO2PXoS",
}
"""Dynamic Secret Salts."""

GEETEST_RETCODES = {10035, 5003, 10041, 1034}
"""API error codes that indicate a Geetest was triggered during the API request."""

APP_KEYS = {
    types.Game.GENSHIN: {
        types.Region.OVERSEAS: "6a4c78fe0356ba4673b8071127b28123",
        types.Region.CHINESE: "d0d3a7342df2026a70f650b907800111",
    },
    types.Game.STARRAIL: {
        types.Region.OVERSEAS: "d74818dabd4182d4fbac7f8df1622648",
        types.Region.CHINESE: "4650f3a396d34d576c3d65df26415394",
    },
    types.Game.HONKAI: {
        types.Region.OVERSEAS: "243187699ab762b682a2a2e50ba02285",
        types.Region.CHINESE: "0ebc517adb1b62c6b408df153331f9aa",
    },
    types.Game.ZZZ: {
        types.Region.OVERSEAS: "ff0f2776bf515d79d1f8ff1fb98b2a06",
        types.Region.CHINESE: "4650f3a396d34d576c3d65df26415394",
    },
}
"""App keys used for game login."""

APP_IDS = {
    types.Game.GENSHIN: {
        types.Region.OVERSEAS: "4",
        types.Region.CHINESE: "4",
    },
    types.Game.STARRAIL: {
        types.Region.OVERSEAS: "11",
        types.Region.CHINESE: "8",
    },
    types.Game.HONKAI: {
        types.Region.OVERSEAS: "8",
        types.Region.CHINESE: "1",
    },
    types.Game.ZZZ: {
        types.Region.OVERSEAS: "15",
        types.Region.CHINESE: "12",
    },
}
"""App IDs used for game login."""

GEETEST_RECORD_KEYS = {
    types.Game.GENSHIN: "genshin_game_record",
    types.Game.STARRAIL: "hkrpg_game_record",
    types.Game.HONKAI: "bh3_game_record",
    types.Game.ZZZ: "nap_game_record",
    types.Game.TOT: "nxx_game_record",
}
"""Keys used to submit geetest result."""

CN_TIMEZONE = datetime.timezone(datetime.timedelta(hours=8))

GAME_BIZS = {
    types.Region.OVERSEAS: {
        types.Game.GENSHIN: "hk4e_global",
        types.Game.STARRAIL: "hkrpg_global",
        types.Game.HONKAI: "bh3_os",
        types.Game.ZZZ: "nap_global",
    },
    types.Region.CHINESE: {
        types.Game.GENSHIN: "hk4e_cn",
        types.Game.STARRAIL: "hkrpg_cn",
        types.Game.HONKAI: "bh3_cn",
        types.Game.ZZZ: "nap_cn",
    },
}

WEB_EVENT_GAME_IDS = {
    types.Game.GENSHIN: 2,
    types.Game.STARRAIL: 6,
    types.Game.HONKAI: 1,
    types.Game.ZZZ: 8,
    types.Game.TOT: 4,
}

# Server names are from uid.py
SERVER_TIMEZONES: typing.Final[dict[int, typing.Tuple[str, ...]]] = {
    8: (
        "cn_gf01",
        "cn_qd01",
        "os_asia",
        "os_cht",
        "prod_gf_cn",
        "prod_qd_cn",
        "prod_official_asia",
        "prod_official_cht",
        "prod_gf_sg",
        "prod_gf_jp",
        "prod_gf_cn",
    ),
    -5: ("os_usa", "prod_official_usa", "prod_gf_us"),
    1: ("os_euro", "prod_official_eur", "prod_gf_eu"),
}

CODE_SUPPORTED_GAMES = {types.Game.GENSHIN, types.Game.ZZZ, types.Game.STARRAIL, types.Game.TOT}
"""Games that support redeeming codes."""



"""Types used in the library."""

import enum
import typing

import typing_extensions

if typing.TYPE_CHECKING:
    from genshin.models.model import Unique

__all__ = ["Game", "Region"]

UniqueT = typing.TypeVar("UniqueT", bound="Unique")


class Region(str, enum.Enum):
    """Region to get data from."""

    OVERSEAS = "os"
    """Applies to all overseas APIs."""

    CHINESE = "cn"
    """Applies to all chinese mainland APIs."""


class Game(str, enum.Enum):
    """Hoyoverse game."""

    GENSHIN = "genshin"
    """Genshin Impact"""

    HONKAI = "honkai3rd"
    """Honkai Impact 3rd"""

    STARRAIL = "hkrpg"
    """Honkai Star Rail"""

    ZZZ = "nap"
    """Zenless Zone Zero"""

    TOT = "tot"
    """Tears of Themis"""


IDOr: typing_extensions.TypeAlias = typing.Union[int, UniqueT]
"""Allows partial objects."""

Lang: typing_extensions.TypeAlias = typing.Literal[
    "zh-cn",
    "zh-tw",
    "de-de",
    "en-us",
    "es-es",
    "fr-fr",
    "id-id",
    "it-it",
    "ja-jp",
    "ko-kr",
    "pt-pt",
    "ru-ru",
    "th-th",
    "vi-vn",
    "tr-tr",
]



"""Email verification -related models"""

import json
import typing

import pydantic

__all__ = [
    "ActionTicket",
    "VerifyStrategy",
]


class VerifyStrategy(pydantic.BaseModel):
    """Verification strategy."""

    ticket: str
    verify_type: str


class ActionTicket(pydantic.BaseModel):
    """Action ticket. Can be used to verify email addresses."""

    risk_ticket: str
    verify_str: VerifyStrategy

    @pydantic.model_validator(mode="before")
    def __parse_data(cls, data: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Parse the data if it was provided in a raw format."""
        verify_str = data["verify_str"]
        if isinstance(verify_str, str):
            data["verify_str"] = json.loads(verify_str)

        return data

    def to_rpc_verify_header(self) -> str:
        """Convert the action ticket to `x-rpc-verify` header."""
        ticket = self.model_dump()
        ticket["verify_str"] = json.dumps(ticket["verify_str"])
        return json.dumps(ticket)



"""Honkai battlesuit model."""

import logging
import re

import pydantic

from genshin.models.model import Aliased, APIModel, Unique

from .constants import BATTLESUIT_IDENTIFIERS

__all__ = ["Battlesuit"]

_LOGGER = logging.getLogger(__name__)

BATTLESUIT_TYPES = {
    "ShengWu": "BIO",
    "JiXie": "MECH",
    "YiNeng": "PSY",
    "LiangZi": "QUA",
    "XuShu": "IMG",
    "Xingchen": "SD",
}
ICON_BASE = "https://upload-os-bbs.mihoyo.com/game_record/honkai3rd/global/SpriteOutput/"


class Battlesuit(APIModel, Unique):
    """Represents a battlesuit without equipment or level."""

    id: int
    name: str
    rarity: int = Aliased("star")
    closeup_icon_background: str = Aliased("avatar_background_path")
    tall_icon: str = Aliased("figure_path")
    banner_art: str = Aliased("image_path")

    @pydantic.field_validator("tall_icon")
    def __autocomplete_figpath(cls, tall_icon: str, info: pydantic.ValidationInfo) -> str:
        """figure_path is empty for gamemode endpoints, and cannot be inferred from other fields."""
        if tall_icon:
            # might as well just update the BATTLESUIT_IDENTIFIERS if we have the data
            if info.data["id"] not in BATTLESUIT_IDENTIFIERS:
                _LOGGER.debug("Updating BATTLESUIT_IDENTIFIERS with %s", tall_icon)
                BATTLESUIT_IDENTIFIERS[info.data["id"]] = tall_icon.split("/")[-1].split(".")[0]

            return tall_icon

        suit_identifier = BATTLESUIT_IDENTIFIERS.get(info.data["id"])
        return ICON_BASE + f"AvatarTachie/{suit_identifier or 'Unknown'}.png"

    @property
    def character(self) -> str:
        match = re.match(r".*/(\w+C\d+).png", self.tall_icon)
        char_raw = match.group(1) if match else ""

        if "Twin" in char_raw:
            # Rozaliya and Liliya share the same identifier
            return {"TwinC1": "Liliya", "TwinC2": "Rozaliya"}[char_raw]

        # fix mislocalized names
        char_name_raw = char_raw.rsplit("C", 1)[0]
        if char_name_raw == "Fuka":
            return "Fu Hua"

        elif char_name_raw == "Lisushang":
            return "Li Sushang"

        return char_name_raw

    @property
    def rank(self) -> str:
        """Display character rarity with letters ranging from A to SSS, as is done in-game."""
        return ("A", "B", "S", "SS", "SSS")[self.rarity - 1]

    @property
    def _type_cn(self) -> str:
        match = re.match(r".*/Attr(\w+?)(?:Small)?.png", self.closeup_icon_background)
        return match[1] if match else "ShengWu"  # random default just so images keep working

    @property
    def type(self) -> str:
        return BATTLESUIT_TYPES[self._type_cn]

    @property
    def closeup_icon(self) -> str:
        return f"{ICON_BASE}AvatarCardIcons/{60000 + self.id}.png"

    @property
    def icon(self) -> str:
        return f"{ICON_BASE}AvatarIcon/{self.id}.png"

    @property
    def icon_background(self) -> str:
        return f"{ICON_BASE}AvatarIcon/Attr{self._type_cn}.png"

    @property
    def image(self) -> str:
        return f"{ICON_BASE}AvatarCardFigures/{60000 + self.id}.png"

    @property
    def cropped_icon(self) -> str:
        return f"{self.image[:-4]}@1.png"

    @property
    def banner(self) -> str:
        return f"{self.image[:-4]}@2.png"



"""Wish component."""

import asyncio
import functools
import typing
import urllib.parse
import warnings

from genshin import paginators, types, utility
from genshin.client import cache as client_cache
from genshin.client import routes
from genshin.client.components import base
from genshin.models.genshin import gacha as models
from genshin.utility import deprecation

__all__ = ["WishClient"]

FATE_BANNER_TYPES = {models.StarRailBannerType.FATE_CHARACTER, models.StarRailBannerType.FATE_WEAPON}
MW_BANNER_TYPES = {
    models.MWBannerType.STANDARD,
    models.MWBannerType.EVENT,
}


class WishClient(base.BaseClient):
    """Wish component."""

    async def request_gacha_info(
        self,
        endpoint: str,
        *,
        lang: typing.Optional[str] = None,
        game: typing.Optional[types.Game] = None,
        authkey: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        short_lang_code: bool = True,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the gacha info endpoint."""
        params = dict(params or {})
        game = game or self.default_game
        if game is None:
            raise RuntimeError("No game provided")

        authkey = authkey or self.authkeys.get(game)
        if authkey is None:
            raise RuntimeError("No authkey provided")

        base_url = routes.GACHA_URL.get_url(self.region, game)
        url = base_url / endpoint

        params["authkey_ver"] = 1
        params["authkey"] = urllib.parse.unquote(authkey)
        params["lang"] = utility.create_short_lang_code(lang or self.lang) if short_lang_code else lang or self.lang
        params["game_biz"] = utility.get_prod_game_biz(self.region, game)

        return await self.request(url, params=params, **kwargs)

    async def _get_gacha_page(
        self,
        end_id: int,
        banner_type: int,
        *,
        game: typing.Optional[types.Game] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> tuple[typing.Sequence[typing.Any], int]:
        """Get a single page of wishes."""
        if banner_type in FATE_BANNER_TYPES:
            endpoint = "getLdGachaLog"
        elif banner_type in MW_BANNER_TYPES:
            endpoint = "getBeyondGachaLog"
        else:
            endpoint = "getGachaLog"

        data = await self.request_gacha_info(
            endpoint,
            lang=lang,
            game=game,
            authkey=authkey,
            params=dict(gacha_type=banner_type, real_gacha_type=banner_type, size=20, end_id=end_id),
            short_lang_code=banner_type not in MW_BANNER_TYPES,
        )

        if game is types.Game.GENSHIN:
            # Genshin doesn't return timezone data
            # America: UTC-5, Europe: UTC+1, others are UTC+8
            tz_offsets = {"os_usa": -13, "os_euro": -7}

            try:
                if isinstance(banner_type, models.MWBannerType):
                    # MW has region inside the wish item
                    tz_offset = tz_offsets.get(data["list"][0]["region"], 0)
                else:
                    tz_offset = tz_offsets.get(data["region"], 0)
            except (KeyError, IndexError):
                tz_offset = 0
        else:
            tz_offset = data["region_time_zone"]
            if game is types.Game.STARRAIL:
                tz_offset -= 8  # Star rail returns UTC+n for this value

        return data["list"], tz_offset

    async def _get_wish_page(
        self,
        end_id: int,
        banner_type: models.GenshinBannerType,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Wish]:
        """Get a single page of wishes."""
        data, tz_offset = await self._get_gacha_page(
            end_id=end_id,
            banner_type=banner_type,
            lang=lang,
            authkey=authkey,
            game=types.Game.GENSHIN,
        )
        return [models.Wish(**i, banner_type=banner_type, tz_offset=tz_offset) for i in data]

    async def _get_mw_wish_page(
        self,
        end_id: int,
        banner_type: models.MWBannerType,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[models.MWWish]:
        """Get a single page of Miliastra Wonderland wishes."""
        data, tz_offset = await self._get_gacha_page(
            end_id=end_id,
            banner_type=banner_type,
            lang=lang,
            authkey=authkey,
            game=types.Game.GENSHIN,
        )
        return [models.MWWish(**i, banner_type=banner_type, tz_offset=tz_offset) for i in data]

    async def _get_warp_page(
        self,
        end_id: int,
        banner_type: models.StarRailBannerType,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Warp]:
        """Get a single page of warps."""
        data, tz_offset = await self._get_gacha_page(
            end_id=end_id,
            banner_type=banner_type,
            lang=lang,
            authkey=authkey,
            game=types.Game.STARRAIL,
        )

        return [models.Warp(**i, banner_type=banner_type, tz_offset=tz_offset) for i in data]

    async def _get_signal_page(
        self,
        end_id: int,
        banner_type: models.ZZZBannerType,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Sequence[models.SignalSearch]:
        """Get a single page of signal searches."""
        data, tz_offset = await self._get_gacha_page(
            end_id=end_id,
            banner_type=banner_type,
            lang=lang,
            authkey=authkey,
            game=types.Game.ZZZ,
        )

        return [models.SignalSearch(**i, banner_type=banner_type, tz_offset=tz_offset) for i in data]

    def wish_history(
        self,
        banner_type: typing.Optional[typing.Union[int, typing.Sequence[int]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[models.Wish]:
        """Get the wish history of a user."""
        banner_types = banner_type or list(models.GenshinBannerType)

        if not isinstance(banner_types, typing.Sequence):
            banner_types = [banner_types]

        iterators: list[paginators.Paginator[models.Wish]] = []
        for banner in banner_types:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_wish_page,
                        banner_type=typing.cast(models.GenshinBannerType, banner),
                        lang=lang,
                        authkey=authkey,
                    ),
                    limit=limit,
                    end_id=end_id,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda wish: wish.time.timestamp())

    def mw_wish_history(
        self,
        banner_type: typing.Optional[typing.Union[int, typing.Sequence[int]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[models.MWWish]:
        """Get the Miliastra Wonderland wish history of a user."""
        banner_types = banner_type or (models.MWBannerType.STANDARD, models.MWBannerType.EVENT)

        if not isinstance(banner_types, typing.Sequence):
            banner_types = [banner_types]

        iterators: list[paginators.Paginator[models.MWWish]] = []
        for banner in banner_types:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_mw_wish_page,
                        banner_type=typing.cast(models.MWBannerType, banner),
                        lang=lang,
                        authkey=authkey,
                    ),
                    limit=limit,
                    end_id=end_id,
                    page_size=5,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda wish: wish.time.timestamp())

    def warp_history(
        self,
        banner_type: typing.Optional[typing.Union[int, typing.Sequence[int]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[models.Warp]:
        """Get the warp history of a user."""
        banner_types = banner_type or list(models.StarRailBannerType)

        if not isinstance(banner_types, typing.Sequence):
            banner_types = [banner_types]

        iterators: list[paginators.Paginator[models.Warp]] = []
        for banner in banner_types:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_warp_page,
                        banner_type=typing.cast(models.StarRailBannerType, banner),
                        lang=lang,
                        authkey=authkey,
                    ),
                    limit=limit,
                    end_id=end_id,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda wish: wish.time.timestamp())

    def signal_history(
        self,
        banner_type: typing.Optional[typing.Union[int, typing.Sequence[int]]] = None,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
        end_id: int = 0,
    ) -> paginators.Paginator[models.SignalSearch]:
        """Get the signal search history of a user."""
        banner_types = banner_type or list(models.ZZZBannerType)

        if not isinstance(banner_types, typing.Sequence):
            banner_types = [banner_types]

        iterators: list[paginators.Paginator[models.SignalSearch]] = []
        for banner in banner_types:
            iterators.append(
                paginators.CursorPaginator(
                    functools.partial(
                        self._get_signal_page,
                        banner_type=typing.cast(models.ZZZBannerType, banner),
                        lang=lang,
                        authkey=authkey,
                    ),
                    limit=limit,
                    end_id=end_id,
                )
            )

        if len(iterators) == 1:
            return iterators[0]

        return paginators.MergedPaginator(iterators, key=lambda wish: wish.time.timestamp())

    @deprecation.deprecated("get_genshin_banner_names")
    async def get_banner_names(
        self,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Mapping[int, str]:
        """Get a list of banner names."""
        return await self.get_genshin_banner_names(lang=lang, authkey=authkey)

    async def get_genshin_banner_names(
        self,
        *,
        lang: typing.Optional[str] = None,
        authkey: typing.Optional[str] = None,
    ) -> typing.Mapping[int, str]:
        """Get a list of banner names."""
        data = await self.request_gacha_info(
            "getConfigList",
            lang=lang,
            game=types.Game.GENSHIN,
            authkey=authkey,
            static_cache=client_cache.cache_key("banner", endpoint="names", lang=lang or self.lang),
        )
        return {int(i["key"]): i["name"] for i in data["gacha_type_list"]}

    async def _get_banner_details(
        self,
        banner_id: str,
        *,
        game: typing.Optional[types.Game] = None,
        lang: typing.Optional[str] = None,
    ) -> models.BannerDetails:
        """Get details of a specific banner using its id."""
        lang = lang or self.lang
        game = game or self.default_game
        if game is None:
            raise RuntimeError("No game provided")
        if game == types.Game.STARRAIL:
            warnings.warn("Banner details for Star Rail are not fully supported.")

        region = "hkrpg" if game == types.Game.STARRAIL else "hk4e"
        server = "prod_official_asia" if game == types.Game.STARRAIL else "os_asia"

        data = await self.request_webstatic(
            f"/gacha_info/{region}/{server}/{banner_id}/{lang}.json",
            cache=client_cache.cache_key("banner", endpoint="details", banner=banner_id, lang=lang),
        )
        return models.BannerDetails(**data, banner_id=banner_id)

    @deprecation.deprecated("get_genshin_banner_ids")
    async def get_banner_ids(self) -> typing.Sequence[str]:
        """Get a list of banner ids.

        Uses the current cn banners.
        """
        return await self.get_genshin_banner_ids()

    async def get_genshin_banner_ids(self) -> typing.Sequence[str]:
        """Get a list of banner ids.

        Uses the current cn banners.
        """

        def process_gacha(data: typing.Mapping[str, typing.Any]) -> str:
            # Temporary fix for 4.5 chronicled wish
            if data["gacha_type"] == 500:
                return "8b10b48c52dd6870f92d72e9963b44bb8968ed2f"
            return data["gacha_id"]

        data = await self.request_webstatic(
            "gacha_info/hk4e/cn_gf01/gacha/list.json",
            region=types.Region.CHINESE,
            cache=client_cache.cache_key("banner", endpoint="ids"),
        )
        return list(map(process_gacha, data["data"]["list"]))

    async def get_banner_details(
        self,
        banner_ids: typing.Optional[typing.Sequence[str]] = None,
        *,
        game: typing.Optional[types.Game] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.BannerDetails]:
        """Get all banner details at once in a batch."""
        game = game or self.default_game
        if game == types.Game.STARRAIL and not banner_ids:
            raise RuntimeError("No banner ids provided for star rail")

        banner_ids = banner_ids or await self.get_genshin_banner_ids()

        coros = (self._get_banner_details(i, lang=lang, game=game) for i in banner_ids)
        data = await asyncio.gather(*coros)
        return list(data)

    @deprecation.deprecated("get_genshin_gacha_items")
    async def get_gacha_items(
        self,
        *,
        server: str = "os_asia",
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.GachaItem]:
        """Get the list of characters and weapons that can be gotten from the gacha."""
        return await self.get_genshin_gacha_items(server=server, lang=lang)

    async def get_genshin_gacha_items(
        self,
        *,
        server: str = "os_asia",
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.GachaItem]:
        """Get the list of characters and weapons that can be gotten from the gacha."""
        raise RuntimeError("This method is currently broken, if you know how to fix it, please open an issue.")
        lang = lang or self.lang
        data = await self.request_webstatic(
            f"/hk4e/gacha_info/{server}/items/{lang}.json",
            cache=client_cache.cache_key("banner", endpoint="items", lang=lang),
        )
        return [models.GachaItem(**i) for i in data]



"""Genshin wish models."""

import datetime
import enum
import re
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "BannerDetailItem",
    "BannerDetails",
    "BannerDetailsUpItem",
    "GachaItem",
    "GenshinBannerType",
    "MWBannerType",
    "MWWish",
    "SignalSearch",
    "StarRailBannerType",
    "Warp",
    "Wish",
    "ZZZBannerType",
]


class GenshinBannerType(enum.IntEnum):
    """Banner types in wish histories."""

    NOVICE = 100
    """Temporary novice banner."""

    STANDARD = PERMANENT = 200
    """Permanent standard banner."""

    CHARACTER = 301
    """Rotating character banner."""

    WEAPON = 302
    """Rotating weapon banner."""

    CHRONICLED = 500
    """Chronicled banner."""

    # these are special cases
    # they exist inside the history but should be counted as the same

    CHARACTER1 = 301
    """Character banner #1."""

    CHARACTER2 = 400
    """Character banner #2."""


class MWBannerType(enum.IntEnum):
    """Banner types in Genshin Miliastra Wonderland."""

    STANDARD = PERMANENT = 1000
    """Permanent standard banner."""

    EVENT = 2000
    """Event banner."""

    EVENT_MALE_OUTFIT1 = 20011
    """Male Manekin outfit banner #1."""

    EVENT_MALE_OUTFIT2 = 20012
    """Male Manekin outfit banner #2."""

    EVENT_FEMALE_OUTFIT1 = 20021
    """Female Manekin outfit banner #1."""

    EVENT_FEMALE_OUTFIT2 = 20022
    """Female Manekin outfit banner #2."""


class StarRailBannerType(enum.IntEnum):
    """Banner types in wish histories."""

    STANDARD = PERMANENT = 1
    """Permanent standard banner."""
    NOVICE = 2
    """Temporary novice banner."""
    CHARACTER = 11
    """Rotating character banner."""
    WEAPON = 12
    """Rotating light cone banner."""

    FATE_CHARACTER = 21
    """FGO collab character banner."""
    FATE_WEAPON = 22
    """FGO collab light cone banner."""


class ZZZBannerType(enum.IntEnum):
    """Banner types in wish histories."""

    STANDARD = PERMANENT = 1
    """Permanent standard banner."""

    CHARACTER = 2
    """Rotating character banner."""

    WEAPON = 3
    """Rotating w-engine banner."""

    BANGBOO = 5
    """Bangboo banner."""

    EXCLUSIVE_RESCREENING = 102
    """Exclusive rescreening banner."""

    REVERBERATION = 103
    """W-Engine reverberation banner."""

    def to_chronicle_type(self) -> str:
        """Get the chronicle type string for this banner type."""
        mapping = {
            ZZZBannerType.STANDARD: "GACHA_TYPE_PERMANENT",
            ZZZBannerType.CHARACTER: "GACHA_TYPE_CHARACTER_UP",
            ZZZBannerType.WEAPON: "GACHA_TYPE_WEAPON_UP",
            ZZZBannerType.BANGBOO: "GACHA_TYPE_BANGBOO",
            ZZZBannerType.EXCLUSIVE_RESCREENING: "GACHA_TYPE_CHARACTER_RETURN",
            ZZZBannerType.REVERBERATION: "GACHA_TYPE_WEAPON_RETURN",
        }
        return mapping[self]


class BaseWish(APIModel, Unique):
    """Base wish model."""

    uid: int
    id: int
    name: str
    rarity: int = Aliased("rank_type")

    tz_offset: int
    """Number of hours from UTC+8."""
    time: datetime.datetime
    """Timezone-aware time of when the wish was made"""

    @pydantic.field_validator("rarity", mode="before")
    def __cast_rarity(cls, v: typing.Any) -> int:
        return int(v)

    @pydantic.field_validator("time", mode="before")
    def __parse_time(cls, v: str, info: pydantic.ValidationInfo) -> datetime.datetime:
        return datetime.datetime.fromisoformat(v).replace(
            tzinfo=datetime.timezone(datetime.timedelta(hours=8 + info.data["tz_offset"]))
        )


class Wish(BaseWish):
    """Wish made on any banner."""

    type: str = Aliased("item_type")
    banner_type: GenshinBannerType

    @pydantic.field_validator("banner_type", mode="before")
    def __cast_banner_type(cls, v: typing.Any) -> int:
        return int(v)


class MWWish(BaseWish):
    """Wish made on Genshin Miliastra Wonderland banner."""

    item_id: int
    type: str = Aliased("item_type")
    is_up: bool
    name: str = Aliased("item_name")

    banner_id: int = Aliased("schedule_id")
    banner_type: MWBannerType = Aliased("op_gacha_type")


class Warp(BaseWish):
    """Warp made on any banner."""

    item_id: int
    type: str = Aliased("item_type")

    banner_type: StarRailBannerType
    banner_id: int = Aliased("gacha_id")

    @pydantic.field_validator("banner_type", mode="before")
    def __cast_banner_type(cls, v: typing.Any) -> int:
        return int(v)


class SignalSearch(BaseWish):
    """Signal Search made on any banner."""

    item_id: int
    type: str = Aliased("item_type")

    banner_type: ZZZBannerType

    @pydantic.field_validator("banner_type", mode="before")
    def __cast_banner_type(cls, v: typing.Any) -> int:
        return int(v)

    @classmethod
    def from_chronicle_data(
        cls, data: typing.Mapping[str, typing.Any], uid: int, tz_offset: int, banner_type: ZZZBannerType
    ) -> "SignalSearch":
        """Create a ZZZChronicleWish from chronicle data."""
        rarity_convert = {"S": 4, "A": 3, "B": 2}
        wish_time = data["date"]
        wish_dt = datetime.datetime(
            year=wish_time["year"],
            month=wish_time["month"],
            day=wish_time["day"],
            hour=wish_time["hour"],
            minute=wish_time["minute"],
            second=wish_time["second"],
        )
        converted_data = {
            "uid": uid,
            "id": int(data["id"]),
            "name": data["item_name"],
            "rank_type": rarity_convert[data["rarity"]],
            "tz_offset": tz_offset,
            "time": wish_dt.isoformat(),
            "item_id": int(data["item_id"]),
            "item_type": data["item_type"],
            "banner_type": banner_type,
        }
        return cls.model_validate(converted_data)


class BannerDetailItem(APIModel):
    """Item that may be gotten from a banner."""

    name: str = Aliased("item_name")
    type: str = Aliased("item_type")
    rarity: int = Aliased("rank")

    up: bool = Aliased("is_up")
    order: int = Aliased("order_value")


class BannerDetailsUpItem(APIModel):
    """Item that has a rate-up on a banner."""

    name: str = Aliased("item_name")
    type: str = Aliased("item_type")
    element: str = Aliased("item_attr")
    icon: str = Aliased("item_img")

    @pydantic.field_validator("element", mode="before")
    def __parse_element(cls, v: str) -> str:
        return {
            "风": "Anemo",
            "火": "Pyro",
            "水": "Hydro",
            "雷": "Electro",
            "冰": "Cryo",
            "岩": "Geo",
            "草": "Dendro",
            "": "",
        }.get(v, v)


class BannerDetails(APIModel):
    """Details of a banner."""

    banner_id: str
    banner_type: int = Aliased("gacha_type")
    title: str
    content: str
    date_range: str

    r5_up_prob: typing.Optional[float]
    r4_up_prob: typing.Optional[float]
    r5_prob: typing.Optional[float]
    r4_prob: typing.Optional[float]
    r3_prob: typing.Optional[float]
    r5_guarantee_prob: typing.Optional[float] = Aliased("r5_baodi_prob")
    r4_guarantee_prob: typing.Optional[float] = Aliased("r4_baodi_prob")
    r3_guarantee_prob: typing.Optional[float] = Aliased("r3_baodi_prob")

    r5_up_items: typing.Sequence[BannerDetailsUpItem]
    r4_up_items: typing.Sequence[BannerDetailsUpItem]

    r5_items: list[BannerDetailItem] = Aliased("r5_prob_list")
    r4_items: list[BannerDetailItem] = Aliased("r4_prob_list")
    r3_items: list[BannerDetailItem] = Aliased("r3_prob_list")

    @pydantic.field_validator("r5_up_items", "r4_up_items", mode="before")
    def __replace_none(cls, v: typing.Optional[typing.Sequence[typing.Any]]) -> typing.Sequence[typing.Any]:
        return v or []

    @pydantic.field_validator(
        "r5_up_prob",
        "r4_up_prob",
        "r5_prob",
        "r4_prob",
        "r3_prob",
        "r5_guarantee_prob",
        "r4_guarantee_prob",
        "r3_guarantee_prob",
        mode="before",
    )
    def __parse_percentage(cls, v: typing.Optional[str]) -> typing.Optional[float]:
        if v is None or isinstance(v, (int, float)):
            return v

        return None if v == "0%" else float(v[:-1].replace(",", "."))

    @property
    def name(self) -> str:
        return re.sub(r"<.*?>", "", self.title).strip()

    @property
    def banner_type_name(self) -> str:
        banners = {
            100: "Novice Wishes",
            200: "Permanent Wish",
            301: "Character Event Wish",
            302: "Weapon Event Wish",
            400: "Character Event Wish",
        }
        return banners[self.banner_type]

    @property
    def items(self) -> typing.Sequence[BannerDetailItem]:
        items = self.r5_items + self.r4_items + self.r3_items
        return sorted(items, key=lambda x: x.order)


class GachaItem(APIModel, Unique):
    """Item that can be gotten from the gacha."""

    name: str
    type: str = Aliased("item_type")
    rarity: int = Aliased("rank_type")
    id: int = Aliased("item_id")

    @pydantic.field_validator("id")
    def __format_id(cls, v: int) -> int:
        return 10000000 + v - 1000 if len(str(v)) == 4 else v

    def is_character(self) -> bool:
        """Whether this item is a character."""
        return len(str(self.id)) == 8



"""Auth-related utility.

Credits to:
- JokelBaf - https://github.com/jokelbaf
- Seria - https://github.com/seriaati
- M-307 - https://github.com/mrwan200
- gsuid_core - https://github.com/Genshin-bots/gsuid_core
"""

from .client import *



"""Honkai stats models."""

import typing

import pydantic

from genshin.models import hoyolab
from genshin.models.model import Aliased, APIModel

from . import battlesuits as honkai_battlesuits
from . import modes

__all__ = ["FullHonkaiUserStats", "HonkaiStats", "HonkaiUserStats"]


# flake8: noqa: E222
class MemorialArenaStats(APIModel):
    """Represents a user's stats regarding the Memorial Arena gamemodes."""

    ranking: float = Aliased("battle_field_ranking_percentage")
    raw_rank: int = Aliased("battle_field_rank")
    score: int = Aliased("battle_field_score")
    raw_tier: int = Aliased("battle_field_area")

    @pydantic.field_validator("ranking", mode="before")
    def __normalize_ranking(cls, value: typing.Union[str, float]) -> float:
        return float(value) if value else 0

    @property
    def rank(self) -> str:
        """The user's Memorial Arena rank as displayed in-game."""
        return modes.prettify_MA_rank(self.raw_rank)


# flake8: noqa: E222
class SuperstringAbyssStats(APIModel):
    """Represents a user's stats regarding Superstring Abyss."""

    raw_rank: int = Aliased("level")
    trophies: int = Aliased("cup_number")
    score: int = Aliased("abyss_score")
    raw_tier: int = Aliased("battle_field_area")

    # for consistency between types; also allows us to forego the mi18n fuckery
    latest_type: typing.ClassVar[str] = "Superstring"


# flake8: noqa: E222
class OldAbyssStats(APIModel):
    """Represents a user's stats regarding Q-Singularis and Dirac Sea."""

    raw_q_singularis_rank: typing.Optional[int] = Aliased("level_of_quantum")
    raw_dirac_sea_rank: typing.Optional[int] = Aliased("level_of_ow")
    score: int = Aliased("abyss_score")
    raw_tier: int = Aliased("latest_area")
    raw_latest_rank: typing.Optional[int] = Aliased("latest_level")
    # TODO: Add proper key
    latest_type: str = Aliased()

    @pydantic.field_validator("raw_q_singularis_rank", "raw_dirac_sea_rank", "raw_latest_rank", mode="before")
    def __normalize_rank(cls, rank: typing.Optional[str]) -> typing.Optional[int]:  # modes.OldAbyss.__normalize_rank
        if isinstance(rank, int):
            return rank

        if rank is None or "Unknown" in rank:
            return None

        return 69 - ord(rank)

    model_config: pydantic.ConfigDict = pydantic.ConfigDict(frozen=False)  # type: ignore


# flake8: noqa: E222
class ElysianRealmStats(APIModel):
    """Represents a user's stats regarding Elysian Realms."""

    highest_difficulty: int = Aliased("god_war_max_punish_level")
    remembrance_sigils: int = Aliased("god_war_extra_item_number")
    highest_score: int = Aliased("god_war_max_challenge_score")
    highest_floor: int = Aliased("god_war_max_challenge_level")
    max_level_suits: int = Aliased("god_war_max_level_avatar_number")


class HonkaiStats(APIModel):
    """Represents a user's stat page"""

    active_days: int = Aliased("active_day_number")
    achievements: int = Aliased("achievement_number")

    battlesuits: int = Aliased("armor_number")
    battlesuits_SSS: int = Aliased("sss_armor_number")
    stigmata: int = Aliased("stigmata_number")
    stigmata_5star: int = Aliased("five_star_stigmata_number")
    weapons: int = Aliased("weapon_number")
    weapons_5star: int = Aliased("five_star_weapon_number")
    outfits: int = Aliased("suit_number")

    abyss: typing.Union[SuperstringAbyssStats, OldAbyssStats] = Aliased()
    memorial_arena: MemorialArenaStats = Aliased()
    elysian_realm: ElysianRealmStats = Aliased()

    @pydantic.model_validator(mode="before")
    def __pack_gamemode_stats(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        if "new_abyss" in values:
            values["abyss"] = SuperstringAbyssStats(**values["new_abyss"], **values)
        elif "old_abyss" in values:
            values["abyss"] = OldAbyssStats(**values["old_abyss"], **values)

        if "memorial_arena" not in values:
            values["memorial_arena"] = MemorialArenaStats(**values)

        if "elysian_realm" not in values:
            values["elysian_realm"] = ElysianRealmStats(**values)

        return values


class HonkaiUserStats(APIModel):
    """Represents basic user stats, showing only generic user data and stats."""

    info: hoyolab.UserInfo = Aliased("role")
    stats: HonkaiStats


class FullHonkaiUserStats(HonkaiUserStats):
    """Represents a user's full stats, including characters, gear, and gamemode data"""

    battlesuits: typing.Sequence[honkai_battlesuits.FullBattlesuit]
    abyss: typing.Sequence[typing.Union[modes.SuperstringAbyss, modes.OldAbyss]]
    memorial_arena: typing.Sequence[modes.MemorialArena]
    elysian_realm: typing.Sequence[modes.ElysianRealm]

    @property
    def abyss_superstring(self) -> typing.Sequence[modes.SuperstringAbyss]:
        """Filter `self.abyss` to only return instances of Superstring Abyss."""
        return [entry for entry in self.abyss if isinstance(entry, modes.SuperstringAbyss)]



"""Event calendar models."""

import typing

import pydantic

from genshin.models.model import Aliased, APIModel, DateTime

__all__ = (
    "AbyssDetail",
    "Banner",
    "BannerCharacter",
    "BannerWeapon",
    "DoubleRewardDetail",
    "Event",
    "EventExplorationDetail",
    "EventReward",
    "GenshinEventCalendar",
    "TheaterDetail",
)


class BannerCharacter(APIModel):
    """Banner character."""

    id: int
    icon: str
    name: str
    element: str
    rarity: typing.Literal[4, 5]


class BannerWeapon(APIModel):
    """Banner weapon."""

    id: int
    icon: str
    rarity: typing.Literal[4, 5]
    name: str
    wiki_url: str


class Banner(APIModel):
    """Banner model."""

    id: int = Aliased("pool_id")
    version: str = Aliased("version_name")  # e.g. 5.2
    name: str = Aliased("pool_name")
    type: int = Aliased("pool_type")

    characters: typing.Sequence[BannerCharacter] = Aliased("avatars")
    weapons: typing.Sequence[BannerWeapon] = Aliased("weapon")

    start_timestamp: int
    end_timestamp: int
    start_time: DateTime
    end_time: DateTime

    jump_url: str
    pool_status: int
    countdown_seconds: int


class EventReward(APIModel):
    """Event reward model."""

    id: int = Aliased("item_id")
    name: str
    icon: str
    wiki_url: str
    num: int
    rarity: int
    homepage_show: bool


class EventExplorationDetail(APIModel):
    """Event exploration detail."""

    explored_percentage: float = Aliased("explore_percent")
    is_finished: bool


class DoubleRewardDetail(APIModel):
    """Double reward detail."""

    total: int
    remaining: int = Aliased("left")


class AbyssDetail(APIModel):
    """Spiral abyss detail."""

    unlocked: bool = Aliased("is_unlock")
    max_star: int
    total_star: int
    has_data: bool


class TheaterDetail(APIModel):
    """Imaginarium theater detail."""

    unlocked: bool = Aliased("is_unlock")
    max_round: int = Aliased("max_round_id")
    has_data: bool


class Event(APIModel):
    """Event model."""

    id: int
    name: str
    description: str = Aliased("desc")
    strategy: str
    type: str

    start_timestamp: int
    end_timestamp: int
    start_time: typing.Optional[DateTime] = None
    end_time: typing.Optional[DateTime] = None

    status: int
    countdown_seconds: int
    rewards: typing.Sequence[EventReward] = Aliased("reward_list")
    is_finished: bool

    exploration_detail: typing.Optional[EventExplorationDetail] = Aliased("explore_detail", default=None)
    double_reward_detail: typing.Optional[DoubleRewardDetail] = Aliased("double_detail", default=None)
    abyss_detail: typing.Optional[AbyssDetail] = Aliased("tower_detail", default=None)
    theater_detail: typing.Optional[TheaterDetail] = Aliased("role_combat_detail", default=None)

    @pydantic.field_validator("description", mode="after")
    def __format_description(cls, v: str) -> str:
        return v.replace("\\n", "\n")


class GenshinEventCalendar(APIModel):
    """Genshin event calendar."""

    character_banners: typing.Sequence[Banner] = Aliased("avatar_card_pool_list")
    weapon_banners: typing.Sequence[Banner] = Aliased("weapon_card_pool_list")
    chronicled_banners: typing.Sequence[Banner] = Aliased("mixed_card_pool_list")
    events: typing.Sequence[Event] = Aliased("act_list")
    challenges: typing.Sequence[Event] = Aliased("fixed_act_list")



from enum import IntEnum

import pydantic

__all__ = ("EnvisagedEchoCharacter", "EnvisagedEchoStatus")


class EnvisagedEchoStatus(IntEnum):
    """Status of an Envisaged Echo character."""

    LOCKED = 1
    UNLOCKED = 2
    CHECKED = 3
    """With a green checkmark icon."""


class EnvisagedEchoCharacter(pydantic.BaseModel):
    """Genshin Impact Envisaged Echo character."""

    id: int = pydantic.Field(alias="avatar_id")
    name: str
    icon: str
    status: EnvisagedEchoStatus
    has_red_dot: bool
    level_id: int



"""Reverse-compatibility layer for previous versions."""

from __future__ import annotations

import typing

import aiohttp

from genshin import models, types
from genshin.utility import deprecation

from . import clients

__all__ = ["ChineseClient", "ChineseMultiCookieClient", "GenshinClient", "MultiCookieClient"]


class GenshinClient(clients.Client):
    """A simple http client for genshin endpoints.

    !!! warning
        This class is deprecated and will be removed in the following version.
        Use `Client()` instead.
    """

    def __init__(
        self,
        cookies: typing.Optional[typing.Any] = None,
        authkey: typing.Optional[str] = None,
        *,
        lang: types.Lang = "en-us",
        region: types.Region = types.Region.OVERSEAS,
        debug: bool = False,
    ) -> None:
        deprecation.warn_deprecated(self.__class__, alternative="Client")
        super().__init__(
            cookies=cookies,
            authkey=authkey,
            lang=lang,
            region=region,
            game=types.Game.GENSHIN,
            debug=debug,
        )

    @property
    def session(self) -> aiohttp.ClientSession:
        """The current client session, created when needed."""
        deprecation.warn_deprecated(self.__class__.session)
        return self.cookie_manager.create_session()

    @property
    def cookies(self) -> typing.Mapping[str, str]:
        """The cookie jar belonging to the current session."""
        deprecation.warn_deprecated(self.__class__.cookies, alternative="cookie_manager")
        return getattr(self.cookie_manager, "cookies")

    @cookies.setter
    def cookies(self, cookies: typing.Mapping[str, typing.Any]) -> None:
        deprecation.warn_deprecated("Setting cookies with GenshinClient.cookies", alternative="set_cookies")
        setattr(self.cookie_manager, "cookies", cookies)

    @property
    def uid(self) -> typing.Optional[int]:
        deprecation.warn_deprecated(self.__class__.uid, alternative="Client.uids[genshin.Game.GENSHIN]")
        return self.uids[types.Game.GENSHIN]

    @uid.setter
    def uid(self, uid: typing.Optional[int]) -> None:
        deprecation.warn_deprecated(
            "Setting uid with GenshinClient.uid",
            alternative="Client.uids[genshin.Game.GENSHIN]",
        )
        if uid is not None:
            self.uids[types.Game.GENSHIN] = uid

    @deprecation.deprecated()
    async def __aenter__(self) -> GenshinClient:
        return self

    async def __aexit__(self, *exc_info: typing.Any) -> None:
        pass

    @deprecation.deprecated("get_partial_genshin_user")
    async def get_partial_user(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.PartialGenshinUserStats:
        """Get partial genshin user without character equipment."""
        return await self.get_partial_genshin_user(uid, lang=lang)

    @deprecation.deprecated("get_genshin_characters")
    async def get_characters(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.Character]:
        """Get genshin user characters."""
        return await self.get_genshin_characters(uid, lang=lang)

    @deprecation.deprecated("get_genshin_user")
    async def get_user(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.GenshinUserStats:
        """Get genshin user."""
        return await self.get_genshin_user(uid, lang=lang)

    @deprecation.deprecated("get_full_genshin_user")
    async def get_full_user(
        self,
        uid: int,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.FullGenshinUserStats:
        """Get a user with all their possible data."""
        return await self.get_full_genshin_user(uid, lang=lang)


class ChineseClient(GenshinClient):
    """A Genshin Client for chinese endpoints.

    !!! warning
        This class is deprecated and will be removed in the following version.
        Use `Client(region=genshin.Region.CHINESE)` instead.
    """

    def __init__(
        self,
        cookies: typing.Optional[typing.Mapping[str, str]] = None,
        authkey: typing.Optional[str] = None,
        *,
        lang: types.Lang = "zh-cn",
        debug: bool = False,
    ) -> None:
        super().__init__(
            cookies=cookies,
            authkey=authkey,
            lang=lang,
            region=types.Region.CHINESE,
            debug=debug,
        )


class MultiCookieClient(GenshinClient):
    """A Genshin Client which allows setting multiple cookies.

    !!! warning
        This class is deprecated and will be removed in the following version.
        Use `Client(cookies=[...])` instead.
    """

    def __init__(
        self,
        cookie_list: typing.Optional[typing.Sequence[typing.Mapping[str, str]]] = None,
        *,
        lang: types.Lang = "en-us",
        debug: bool = False,
    ) -> None:
        super().__init__(
            cookies=cookie_list,
            lang=lang,
            debug=debug,
        )


class ChineseMultiCookieClient(GenshinClient):
    """A Genshin Client for chinese endpoints which allows setting multiple cookies.

    !!! warning
        This class is deprecated and will be removed in the following version.
        Use `Client(cookies=[...], region=genshin.Region.CHINESE)` instead.
    """

    def __init__(
        self,
        cookie_list: typing.Optional[typing.Sequence[typing.Mapping[str, str]]] = None,
        *,
        lang: types.Lang = "en-us",
        debug: bool = False,
    ) -> None:
        super().__init__(
            cookies=cookie_list,
            lang=lang,
            region=types.Region.CHINESE,
            debug=debug,
        )



import datetime
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, DateTime, TZDateTime
from genshin.models.zzz.character import ZZZElementType, ZZZSpecialty

from .common import ChallengeBangboo

__all__ = (
    "DeadlyAssault",
    "DeadlyAssaultAgent",
    "DeadlyAssaultBoss",
    "DeadlyAssaultBuff",
    "DeadlyAssaultChallenge",
)


class DeadlyAssaultBoss(APIModel):
    """ZZZ Deadly Assault boss."""

    icon: str
    name: str
    background: str = Aliased("bg_icon")
    badge_icon: str = Aliased("race_icon")


class DeadlyAssaultBuff(APIModel):
    """ZZZ Deadly Assault buff model."""

    name: str
    description: str = Aliased("desc")
    icon: str


class DeadlyAssaultAgent(APIModel):
    """ZZZ Deadly Assault agent model."""

    id: int
    level: int
    element: ZZZElementType = Aliased("element_type")
    specialty: ZZZSpecialty = Aliased("avatar_profession")
    rarity: typing.Literal["S", "A"]
    mindscape: int = Aliased("rank")
    icon: str = Aliased("role_square_url")


class DeadlyAssaultChallenge(APIModel):
    """ZZZ Deadly Assault challenge model."""

    score: int
    star: int
    total_star: int
    challenge_time: datetime.datetime

    boss: DeadlyAssaultBoss
    buffs: typing.Sequence[DeadlyAssaultBuff] = Aliased("buffer")
    agents: typing.Sequence[DeadlyAssaultAgent] = Aliased("avatar_list")
    bangboo: typing.Optional[ChallengeBangboo] = Aliased("buddy", default=None)

    @pydantic.field_validator("challenge_time", mode="before")
    def __parse_datetime(cls, value: typing.Mapping[str, typing.Any]) -> typing.Optional[TZDateTime]:
        if value:
            return datetime.datetime(**value)
        return None

    @pydantic.field_validator("boss", mode="before")
    def __parse_boss(cls, value: typing.List[typing.Mapping[str, typing.Any]]) -> DeadlyAssaultBoss:
        if not value:
            raise ValueError("No boss data provided.")
        return DeadlyAssaultBoss(**value[0])


class DeadlyAssault(APIModel):
    """ZZZ Deadly Assault model."""

    id: int = Aliased("zone_id")
    start_time: typing.Optional[DateTime]
    end_time: typing.Optional[DateTime]

    challenges: typing.Sequence[DeadlyAssaultChallenge] = Aliased("list")
    has_data: bool
    total_score: int
    total_star: int
    rank_percent: str

    nickname: str = Aliased("nick_name")
    player_avatar: str = Aliased("avatar_icon")

    @pydantic.field_validator("rank_percent", mode="before")
    def __parse_rank_percent(cls, value: int) -> str:
        return f"{value / 100}%"



"""Base ABC Client."""

import abc
import base64
import functools
import json
import logging
import os
import typing
import urllib.parse
import warnings

import aiohttp.typedefs
import multidict
import yarl

from genshin import constants, errors, types, utility
from genshin.client import cache as client_cache
from genshin.client import routes
from genshin.client.manager import managers
from genshin.models import hoyolab as hoyolab_models
from genshin.utility import concurrency, deprecation, ds
from genshin.utility.uid import recognize_server

__all__ = ["BaseClient"]


T = typing.TypeVar("T")
CallableT = typing.TypeVar("CallableT", bound="typing.Callable[..., object]")
AsyncCallableT = typing.TypeVar("AsyncCallableT", bound="typing.Callable[..., typing.Awaitable[object]]")


def parse_loose_headers(
    loose_headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
) -> multidict.CIMultiDict[str]:
    """Parse loose aiohttp headers."""
    return multidict.CIMultiDict((str(k), str(v)) for k, v in dict(loose_headers or ()).items())


class BaseClient(abc.ABC):
    """Base ABC Client."""

    __slots__ = (
        "cookie_manager",
        "cache",
        "_lang",
        "_region",
        "_default_game",
        "uids",
        "authkeys",
        "_hoyolab_id",
        "_accounts",
        "custom_headers",
    )

    USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36"  # noqa: E501

    logger: logging.Logger = logging.getLogger(__name__)

    cookie_manager: managers.BaseCookieManager
    cache: client_cache.BaseCache
    _lang: str
    _region: types.Region
    _default_game: typing.Optional[types.Game]

    uids: dict[types.Game, int]
    authkeys: dict[types.Game, str]
    _hoyolab_id: typing.Optional[int]
    _accounts: dict[types.Game, hoyolab_models.GenshinAccount]
    custom_headers: multidict.CIMultiDict[str]

    def __init__(
        self,
        cookies: typing.Optional[managers.AnyCookieOrHeader] = None,
        *,
        authkey: typing.Optional[str] = None,
        lang: types.Lang = "en-us",
        region: types.Region = types.Region.OVERSEAS,
        proxy: typing.Optional[str] = None,
        game: typing.Optional[types.Game] = None,
        uid: typing.Optional[int] = None,
        hoyolab_id: typing.Optional[int] = None,
        device_id: typing.Optional[str] = None,
        device_fp: typing.Optional[str] = None,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        cache: typing.Optional[client_cache.BaseCache] = None,
        debug: bool = False,
    ) -> None:
        self.cookie_manager = managers.BaseCookieManager.from_cookies(cookies)
        self.cache = cache or client_cache.StaticCache()

        self.uids = {}
        self.authkeys = {}
        self._accounts = {}

        self.default_game = game
        self.lang = lang
        self.region = region
        self.authkey = authkey
        self.debug = debug
        self.proxy = proxy
        self.uid = uid
        self.hoyolab_id = hoyolab_id

        self.custom_headers = parse_loose_headers(headers)
        self.custom_headers.update({"x-rpc-device_id": device_id} if device_id else {})
        self.custom_headers.update({"x-rpc-device_fp": device_fp} if device_fp else {})

    def __repr__(self) -> str:
        kwargs = dict(
            lang=self.lang,
            region=self.region.value,
            default_game=self.default_game and self.default_game.value,
            hoyolab_id=self.hoyolab_id,
            uid=self.default_game and self.uid,
            authkey=self.authkey and self.authkey[:12] + "...",
            proxy=self.proxy,
            debug=self.debug,
        )
        return f"<{type(self).__name__} {', '.join(f'{k}={v!r}' for k, v in kwargs.items() if v)}>"

    @property
    def device_id(self) -> typing.Optional[str]:
        """The device id used in headers."""
        return self.custom_headers.get("x-rpc-device_id")

    @device_id.setter
    def device_id(self, device_id: str) -> None:
        self.custom_headers["x-rpc-device_id"] = device_id

    @property
    def device_fp(self) -> typing.Optional[str]:
        """The device fingerprint used in headers."""
        return self.custom_headers.get("x-rpc-device_fp")

    @device_fp.setter
    def device_fp(self, device_fp: str) -> None:
        self.custom_headers["x-rpc-device_fp"] = device_fp

    @property
    def hoyolab_id(self) -> typing.Optional[int]:
        """The logged-in user's hoyolab uid.

        Returns None if not found or not applicable.
        """
        return self._hoyolab_id or self.cookie_manager.user_id

    @hoyolab_id.setter
    def hoyolab_id(self, hoyolab_id: typing.Optional[int]) -> None:
        if hoyolab_id is None:
            self._hoyolab_id = None
            return

        if self.cookie_manager.multi:
            raise RuntimeError("Cannot specify a hoyolab uid when using multiple cookies.")

        if self.cookie_manager.user_id and hoyolab_id and self.cookie_manager.user_id != hoyolab_id:
            raise ValueError("The provided hoyolab uid does not match the cookie id.")

        self._hoyolab_id = hoyolab_id

    @property
    def lang(self) -> str:
        """The default language, defaults to "en-us" """
        return self._lang

    @lang.setter
    def lang(self, lang: str) -> None:
        if lang not in constants.LANGS:
            raise ValueError(f"{lang} is not a valid language, must be one of: " + ", ".join(constants.LANGS))

        self._lang = lang

    @property
    def region(self) -> types.Region:
        """The default region."""
        return self._region

    @region.setter
    def region(self, region: str) -> None:
        self._region = types.Region(region)

        if region == types.Region.CHINESE:
            self.lang = "zh-cn"

    @property
    def default_game(self) -> typing.Optional[types.Game]:
        """The default game."""
        return self._default_game

    @default_game.setter
    def default_game(self, game: typing.Optional[str]) -> None:
        self._default_game = types.Game(game) if game else None

    game = default_game

    @property
    def uid(self) -> typing.Optional[int]:
        """UID of the default game."""
        if self.default_game is None:
            if len(self.uids) != 1:
                return None

            (self.default_game,) = self.uids.keys()

        return self.uids.get(self.default_game)

    @uid.setter
    def uid(self, uid: typing.Optional[int]) -> None:
        if uid is None:
            self.uids.clear()
            return

        self._default_game = self._default_game or utility.recognize_game(uid, region=self.region)
        if self.default_game is None:
            raise RuntimeError("No default game set. Cannot set uid.")

        self.uids[self.default_game] = uid

    @property
    def authkey(self) -> typing.Optional[str]:
        """The default genshin authkey used for paginators."""
        if self.default_game is None:
            if self.authkeys:
                warnings.warn("Tried to get an authkey without a default game set.")

            return None

        return self.authkeys.get(self.default_game)

    @authkey.setter
    def authkey(self, authkey: typing.Optional[str]) -> None:
        if authkey is None:
            self.authkeys.clear()
            return

        authkey = urllib.parse.unquote(authkey)

        try:
            base64.b64decode(authkey, validate=True)
        except Exception as e:
            raise ValueError("authkey is not a valid base64 encoded string") from e

        if not self.default_game:
            raise RuntimeError("No default game set. Cannot set authkey with property.")

        self.authkeys[self.default_game] = authkey

    @property
    def debug(self) -> bool:
        """Whether the debug logs are being shown in stdout"""
        return logging.getLogger("genshin").level == logging.DEBUG

    @debug.setter
    def debug(self, debug: bool) -> None:
        logging.basicConfig()
        level = logging.DEBUG if debug else logging.NOTSET
        logging.getLogger("genshin").setLevel(level)

    def set_cookies(self, cookies: typing.Optional[managers.AnyCookieOrHeader] = None, **kwargs: typing.Any) -> None:
        """Parse and set cookies."""
        if not bool(cookies) ^ bool(kwargs):
            raise TypeError("Cannot use both positional and keyword arguments at once")

        self.cookie_manager = managers.BaseCookieManager.from_cookies(cookies or kwargs)

    def set_browser_cookies(self, browser: typing.Optional[str] = None) -> None:
        """Extract cookies from your browser and set them as client cookies.

        Available browsers: chrome, chromium, opera, edge, firefox.
        """
        self.cookie_manager = managers.BaseCookieManager.from_browser_cookies(browser)

    def set_authkey(self, authkey: typing.Optional[str] = None, *, game: typing.Optional[types.Game] = None) -> None:
        """Set an authkey for wish & transaction logs.

        Accepts an authkey, a url containing an authkey or a path towards a logfile.
        """
        if authkey is None or os.path.isfile(authkey):
            authkey = utility.get_authkey(authkey)
        else:
            authkey = utility.extract_authkey(authkey) or authkey

        game = game or self.default_game
        if game is None:
            raise RuntimeError("No default game set.")

        self.authkeys[game] = authkey

    def set_cache(
        self, maxsize: int = 1024, *, ttl: int = client_cache.HOUR, static_ttl: int = client_cache.DAY
    ) -> None:
        """Create and set a new cache."""
        self.cache = client_cache.Cache(maxsize, ttl=ttl, static_ttl=static_ttl)

    def set_redis_cache(
        self, url: str, *, ttl: int = client_cache.HOUR, static_ttl: int = client_cache.DAY, **redis_kwargs: typing.Any
    ) -> None:
        """Create and set a new redis cache."""
        import aioredis

        redis = aioredis.Redis.from_url(url, **redis_kwargs)  # pyright: ignore[reportUnknownMemberType]
        self.cache = client_cache.RedisCache(redis, ttl=ttl, static_ttl=static_ttl)

    @property
    def proxy(self) -> typing.Optional[str]:
        """Proxy for http requests."""
        if self.cookie_manager.proxy is None:
            return None

        return str(self.cookie_manager.proxy)

    @proxy.setter
    def proxy(self, proxy: typing.Optional[aiohttp.typedefs.StrOrURL]) -> None:
        self.cookie_manager.proxy = yarl.URL(proxy) if proxy else None

    async def _request_hook(
        self,
        method: str,
        url: aiohttp.typedefs.StrOrURL,
        *,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Any = None,
        **kwargs: typing.Any,
    ) -> None:
        """Perform an action before a request.

        Debug logging by default.
        """
        url = yarl.URL(url)
        if params:
            params = {k: v for k, v in params.items() if k != "authkey"}
            url = url.update_query(params)

        if data:
            self.logger.debug("%s %s\n%s", method, url, json.dumps(data, separators=(",", ":")))
        else:
            self.logger.debug("%s %s", method, url)

    async def request(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        method: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Any = None,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        cache: typing.Any = None,
        static_cache: typing.Any = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request and return a parsed json response."""
        if cache is not None:
            value = await self.cache.get(cache)
            if value is not None:
                return value
        elif static_cache is not None:
            value = await self.cache.get_static(static_cache)
            if value is not None:
                return value

        # actual request

        headers = parse_loose_headers(headers)
        headers["User-Agent"] = self.USER_AGENT
        headers.update(self.custom_headers)

        if method is None:
            method = "POST" if data else "GET"

        if "json" in kwargs:
            raise TypeError("Use data instead of json in request.")

        await self._request_hook(method, url, params=params, data=data, headers=headers, **kwargs)

        response = await self.cookie_manager.request(
            url, method=method, params=params, json=data, headers=headers, **kwargs
        )

        # cache

        if cache is not None:
            await self.cache.set(cache, response)
        elif static_cache is not None:
            await self.cache.set_static(static_cache, response)

        return response

    async def request_webstatic(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        cache: typing.Any = None,
        region: types.Region = types.Region.OVERSEAS,
        **kwargs: typing.Any,
    ) -> typing.Any:
        """Request a static json file."""
        if cache is not None:
            value = await self.cache.get_static(cache)
            if value is not None:
                return value

        url = routes.WEBSTATIC_URL.get_url(region).join(yarl.URL(url))

        headers = parse_loose_headers(headers)
        headers["User-Agent"] = self.USER_AGENT
        headers.update(self.custom_headers)

        await self._request_hook("GET", url, headers=headers, **kwargs)

        async with self.cookie_manager.create_session() as session:
            async with session.get(url, headers=headers, proxy=self.proxy, **kwargs) as r:
                r.raise_for_status()
                data = await r.json()

        if cache is not None:
            await self.cache.set_static(cache, data)

        return data

    async def request_bbs(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        lang: typing.Optional[str] = None,
        region: typing.Optional[types.Region] = None,
        method: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Any = None,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request any bbs endpoint."""
        if lang is not None and lang not in constants.LANGS:
            raise ValueError(f"{lang} is not a valid language, must be one of: " + ", ".join(constants.LANGS))

        lang = lang or self.lang
        region = region or self.region

        url = routes.BBS_URL.get_url(region).join(yarl.URL(url))

        headers = parse_loose_headers(headers)
        headers.update(ds.get_ds_headers(data=data, params=params, region=region, lang=lang or self.lang))
        headers["Referer"] = str(routes.BBS_REFERER_URL.get_url(self.region))

        data = await self.request(url, method=method, params=params, data=data, headers=headers, **kwargs)
        return data

    async def request_hoyolab(
        self,
        url: aiohttp.typedefs.StrOrURL,
        *,
        lang: typing.Optional[str] = None,
        region: typing.Optional[types.Region] = None,
        method: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        data: typing.Any = None,
        headers: typing.Optional[aiohttp.typedefs.LooseHeaders] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request any hoyolab endpoint."""
        if lang is not None and lang not in constants.LANGS:
            raise ValueError(f"{lang} is not a valid language, must be one of: " + ", ".join(constants.LANGS))

        lang = lang or self.lang
        region = region or self.region

        url = routes.TAKUMI_URL.get_url(region).join(yarl.URL(url))

        headers = parse_loose_headers(headers)
        headers.update(ds.get_ds_headers(data=data, params=params, region=region, lang=lang or self.lang))

        data = await self.request(url, method=method, params=params, data=data, headers=headers, **kwargs)
        return data

    @managers.no_multi
    async def get_game_accounts(
        self, *, lang: typing.Optional[str] = None
    ) -> typing.Sequence[hoyolab_models.GenshinAccount]:
        """Get the game accounts of the currently logged-in user."""
        if self.hoyolab_id is None:
            warnings.warn("No hoyolab id set, caching may be unreliable.")

        data = await self.request_hoyolab(
            "binding/api/getUserGameRolesByCookie",
            lang=lang,
            cache=client_cache.cache_key("accounts", hoyolab_id=self.hoyolab_id),
        )
        return [hoyolab_models.GenshinAccount(**i) for i in data["list"]]

    @deprecation.deprecated("get_game_accounts")
    async def genshin_accounts(
        self, *, lang: typing.Optional[str] = None
    ) -> typing.Sequence[hoyolab_models.GenshinAccount]:
        """Get the genshin accounts of the currently logged-in user."""
        accounts = await self.get_game_accounts(lang=lang)
        return [account for account in accounts if account.game == types.Game.GENSHIN]

    async def _update_cached_uids(self) -> None:
        """Update cached fallback uids."""
        mixed_accounts = await self.get_game_accounts()

        game_accounts: dict[types.Game, list[hoyolab_models.GenshinAccount]] = {}
        for account in mixed_accounts:
            if not isinstance(account.game, types.Game):  # pyright: ignore[reportUnnecessaryIsInstance]
                continue

            game_accounts.setdefault(account.game, []).append(account)

        self.uids = {game: max(accounts, key=lambda a: a.level).uid for game, accounts in game_accounts.items()}

        if len(self.uids) == 1 and self.default_game is None:
            (self.default_game,) = self.uids.keys()

    @concurrency.prevent_concurrency
    async def _get_uid(self, game: types.Game) -> int:
        """Get a cached fallback uid."""
        # TODO: use lock
        if uid := self.uids.get(game):
            return uid

        if self.cookie_manager.multi:
            raise RuntimeError("UID must be provided when using multi-cookie managers.")

        await self._update_cached_uids()

        if uid := self.uids.get(game):
            return uid

        raise errors.AccountNotFound(msg="No UID provided and account has no game account bound to it.")

    async def _update_cached_accounts(self) -> None:
        """Update cached fallback accounts."""
        mixed_accounts = await self.get_game_accounts()

        game_accounts: dict[types.Game, list[hoyolab_models.GenshinAccount]] = {}
        for account in mixed_accounts:
            if not isinstance(account.game, types.Game):  # pyright: ignore[reportUnnecessaryIsInstance]
                continue

            game_accounts.setdefault(account.game, []).append(account)

        self._accounts = {}
        for game, accounts in game_accounts.items():
            self._accounts[game] = next(
                (acc for acc in accounts if acc.uid == self.uids.get(game)), max(accounts, key=lambda a: a.level)
            )

    @concurrency.prevent_concurrency
    async def _get_account(self, game: types.Game) -> hoyolab_models.GenshinAccount:
        """Get a cached fallback account."""
        if (account := self._accounts.get(game)) and (uid := self.uids.get(game)) and account.uid == uid:
            return account

        await self._update_cached_accounts()

        if account := self._accounts.get(game):
            if (uid := self.uids.get(game)) and account.uid != uid:
                raise errors.AccountNotFound(msg="There is no game account with such UID.")

            return account

        raise errors.AccountNotFound(msg="Account has no game account bound to it.")

    def _get_hoyolab_id(self) -> int:
        """Get a cached fallback hoyolab ID."""
        if self.hoyolab_id is not None:
            return self.hoyolab_id

        if self.cookie_manager.multi:
            raise RuntimeError("Hoyolab ID must be provided when using multi-cookie managers.")

        raise RuntimeError("No default hoyolab ID provided.")

    def get_account_timezone(
        self, *, game: typing.Optional[types.Game] = None, uid: typing.Optional[int] = None
    ) -> typing.Optional[int]:
        """Get the UTC timezone of the default game account."""
        game = game or self.game
        if game is None:
            return None

        uid = uid or self.uid
        if uid is None:
            return None

        server = recognize_server(uid, game=game)  # type: ignore[arg-type]

        for tz, servers in constants.SERVER_TIMEZONES.items():
            if server in servers:
                return tz

        return None

    def _add_timezone_to_data(
        self,
        data: typing.Mapping[str, typing.Any],
        keys: typing.Sequence[str],
        *,
        game: typing.Optional[types.Game] = None,
        uid: typing.Optional[int] = None,
    ) -> typing.Mapping[str, typing.Any]:
        """Add timezone info to a data dict based on the default game account."""
        tz = self.get_account_timezone(game=game, uid=uid)
        if tz is not None:
            data = dict(data)
            for key in keys:
                if key in data and isinstance(data[key], dict):
                    data[key]["tzinfo"] = tz

        return data


def region_specific(region: types.Region) -> typing.Callable[[AsyncCallableT], AsyncCallableT]:
    """Prevent function to be ran with unsupported regions."""

    def decorator(func: AsyncCallableT) -> AsyncCallableT:
        @functools.wraps(func)
        async def wrapper(self: typing.Any, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            if not hasattr(self, "region"):
                raise TypeError("Cannot use @region_specific on a plain function.")
            if region != self.region:
                raise RuntimeError("The method can only be used with client region set to " + region)

            return await func(self, *args, **kwargs)

        return typing.cast("AsyncCallableT", wrapper)

    return decorator



"""Genshin serenitea pot replica display models."""

from __future__ import annotations

import enum
import typing

import pydantic

from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "TCGBaseCard",
    "TCGCard",
    "TCGCardType",
    "TCGCharacterCard",
    "TCGCharacterTalent",
    "TCGCost",
    "TCGPartialCard",
    "TCGPreview",
]


class TCGCardType(str, enum.Enum):
    """TCG card type."""

    CHARACTER = "CardTypeCharacter"
    EQUIPMENT = "CardTypeModify"
    ASSIST = "CardTypeAssist"
    EVENT = "CardTypeEvent"


class TCGPartialCard(APIModel, Unique):
    """Partial TCG card data."""

    id: int
    image: str


class TCGPreview(APIModel):
    """Preview of TCG stats."""

    character_cards: int = Aliased("avatar_card_num_gained")
    action_cards: int = Aliased("action_card_num_gained")

    total_character_cards: int = Aliased("avatar_card_num_total")
    total_action_cards: int = Aliased("action_card_num_total")

    level: int
    nickname: str

    cards: typing.Sequence[TCGPartialCard] = Aliased("covers")


class TCGCharacterTalent(APIModel):
    """TCG character talent."""

    id: int
    name: str
    description: str = Aliased("desc")
    type: str = Aliased("tag")


class TCGCost(APIModel):
    """TCG cost."""

    element: str = Aliased("cost_type")
    value: int = Aliased("cost_value")

    @pydantic.field_validator("element")
    def __fix_element(cls, value: str) -> str:
        return {
            "CostTypeCryo": "Cryo",
            "CostTypeDendro": "Dendro",
            "CostTypeElectro": "Electro",
            "CostTypeGeo": "Geo",
            "CostTypeHydro": "Hydro",
            "CostTypePyro": "Pyro",
            "CostTypeAnemo": "Anemo",
            "CostTypeSame": "Same",
            "CostTypeVoid": "Void",
        }.get(value, value)


class TCGBaseCard(TCGPartialCard):
    """TCG card."""

    type: TCGCardType = Aliased("card_type")

    name: str
    proficiency: int
    rank_id: int
    usages: int = Aliased("use_count")

    image_tags: typing.Sequence[str] = Aliased("tags")

    wiki_url: str = Aliased("card_wiki")


class TCGCharacterCard(TCGBaseCard):
    """TCG character card."""

    type: typing.Literal[TCGCardType.CHARACTER] = Aliased("card_type")

    name: str
    health: int = Aliased("hp")

    talents: typing.Sequence[TCGCharacterTalent] = Aliased("card_skills")


class TCGCard(TCGBaseCard):
    """TCG equipment card."""

    type: typing.Literal[TCGCardType.EQUIPMENT, TCGCardType.ASSIST, TCGCardType.EVENT] = Aliased("card_type")

    cost: typing.Sequence[TCGCost] = Aliased("action_cost")
    description: str = Aliased("desc")



"""Lineup component."""

import functools
import typing

import genshin.models.genshin as genshin_models
from genshin import paginators, types, utility
from genshin.client import cache, routes
from genshin.client.components import base
from genshin.client.manager import managers
from genshin.models.genshin import lineup as models

__all__ = ["LineupClient"]


class LineupClient(base.BaseClient):
    """Lineup component."""

    async def request_lineup(
        self,
        endpoint: str,
        *,
        method: str = "GET",
        lang: typing.Optional[str] = None,
        params: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        """Make a request towards the lineup endpoint."""
        params = dict(params or {})

        base_url = routes.LINEUP_URL.get_url(self.region)
        url = base_url / endpoint

        params["lang"] = lang or self.lang

        return await self.request(url, method=method, params=params, **kwargs)

    async def get_lineup_fields(
        self, *, lang: typing.Optional[str] = None, use_cache: bool = True
    ) -> models.LineupFields:
        """Get configuration lineup fields."""
        data = await self.request_lineup(
            "config",
            lang=lang,
            static_cache=cache.cache_key("lineup", endpoint="config", lang=lang or self.lang) if use_cache else None,
        )

        return models.LineupFields(**data)

    async def get_lineup_scenarios(
        self,
        *,
        lang: typing.Optional[str] = None,
    ) -> models.LineupScenarios:
        """Get lineup scenarios."""
        data = await self.request_lineup(
            "tags",
            lang=lang,
            static_cache=cache.cache_key("lineup", endpoint="tags", lang=lang or self.lang),
        )
        dummy: dict[str, typing.Any] = dict(id=0, name="", children=data["tree"])

        return models.LineupScenarios(**dummy)

    async def _get_lineup_page(
        self,
        token: str,
        *,
        limit: typing.Optional[int] = None,
        tag_id: typing.Optional[int] = None,
        roles: typing.Optional[typing.Sequence[int]] = None,
        order: typing.Optional[str] = None,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> tuple[str, typing.Sequence[models.LineupPreview]]:
        """Get a single page of lineups."""
        params: dict[str, typing.Any] = dict(
            next_page_token=token,
            limit=limit or "",
            tag_id=tag_id or "",
            order=order or "",
            roles=roles or "",
        )

        if order == "Match":
            uid = uid or await self._get_uid(types.Game.GENSHIN)

            params["uid"] = uid
            params["region"] = utility.recognize_genshin_server(uid)

        data = await self.request_lineup("lineup/index", lang=lang, params=params)

        return data["next_page_token"], [models.LineupPreview(**i) for i in data["list"]]

    def get_lineups(
        self,
        scenario: typing.Optional[types.IDOr[models.LineupScenario]] = None,
        *,
        limit: typing.Optional[int] = None,
        page_size: typing.Optional[int] = None,
        newest: bool = False,
        match_characters: bool = False,
        characters: typing.Optional[typing.Sequence[types.IDOr[genshin_models.Character]]] = None,
        uid: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> paginators.TokenPaginator[models.LineupPreview]:
        """Get lineups."""
        if scenario is not None:
            scenario = int(scenario)

        character_ids = [int(i) for i in characters] if characters is not None else None

        if match_characters:
            order = "Match"
        else:
            order = "CreatedTime" if newest else "Hot"

        return paginators.TokenPaginator(
            functools.partial(
                self._get_lineup_page,
                tag_id=scenario,
                roles=character_ids,
                order=order,
                uid=uid,
                lang=lang,
            ),
            limit=limit,
            page_size=page_size,
        )

    async def get_lineup_details(
        self,
        lineup: typing.Union[str, models.LineupPreview],
        *,
        lang: typing.Optional[str] = None,
    ) -> models.Lineup:
        """Get lineup with detailed characters."""
        lineup_id = lineup if isinstance(lineup, str) else lineup.id

        data = await self.request_lineup(
            "lineup/detail",
            lang=lang,
            params=dict(id=lineup_id),
            cache=cache.cache_key("lineup", endpoint="detail", id=lineup_id, lang=lang or self.lang),
        )

        return models.Lineup(**data["lineup"])

    @managers.no_multi
    async def get_user_lineups(
        self,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.LineupPreview]:
        """Get lineups of the currently logged-in user."""
        data = await self.request_lineup(
            "user/lineup",
            lang=lang,
            params=dict(limit=limit or 1000),
            cache=cache.cache_key("lineup", endpoint="user", lang=lang or self.lang),
        )

        return [models.LineupPreview(**i) for i in data["list"]]

    @managers.no_multi
    async def get_favorite_lineups(
        self,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.LineupPreview]:
        """Get favorited lineups of the currently logged-in user."""
        data = await self.request_lineup(
            "user/favour_lineup",
            lang=lang,
            params=dict(limit=limit or 1000),
            cache=cache.cache_key("lineup", endpoint="favorite", lang=lang or self.lang),
        )

        return [models.LineupPreview(**i) for i in data["list"]]

    @managers.no_multi
    async def get_lineup_character_history(
        self,
        *,
        limit: typing.Optional[int] = None,
        lang: typing.Optional[str] = None,
    ) -> typing.Sequence[models.LineupCharacter]:
        """Get previous character builds of the currently logged-in user.."""
        data = await self.request_lineup(
            "lineup/history",
            lang=lang,
            params=dict(limit=limit or 1000),
            cache=cache.cache_key("lineup", endpoint="history", lang=lang or self.lang),
        )

        return [models.LineupCharacter(**i) for i in data["list"]]



"""Genshin chronicle character."""

import enum
import typing

import pydantic

from genshin.models.genshin import character
from genshin.models.model import Aliased, APIModel, Unique

__all__ = [
    "Artifact",
    "ArtifactProperty",
    "ArtifactSet",
    "ArtifactSetEffect",
    "Character",
    "CharacterSkill",
    "CharacterWeapon",
    "Constellation",
    "DetailArtifact",
    "DetailCharacterWeapon",
    "GenshinDetailCharacter",
    "GenshinDetailCharacters",
    "GenshinWeaponType",
    "Outfit",
    "PartialCharacter",
    "PropInfo",
    "PropertyValue",
    "SkillAffix",
]


class GenshinWeaponType(enum.IntEnum):
    """Character weapon types."""

    SWORD = 1
    CATALYST = 10
    CLAYMORE = 11
    BOW = 12
    POLEARM = 13


class PartialCharacter(character.BaseCharacter):
    """Character without any equipment."""

    level: int
    friendship: int = Aliased("fetter")
    constellation: int = Aliased("actived_constellation_num")


class CharacterWeapon(APIModel, Unique):
    """Character's equipped weapon."""

    id: int
    icon: str
    name: str
    rarity: int
    level: int
    type: int
    refinement: int = Aliased("affix_level")


class ArtifactSetEffect(APIModel):
    """Effect of an artifact set."""

    required_piece_num: int = Aliased("activation_number")
    effect: str
    active: bool = Aliased("enabled", default=False)


class ArtifactSet(APIModel, Unique):
    """Artifact set."""

    id: int
    name: str
    effects: typing.Sequence[ArtifactSetEffect] = Aliased("affixes")


class Artifact(APIModel, Unique):
    """Character's equipped artifact."""

    id: int
    icon: str
    name: str
    pos_name: str
    pos: int
    rarity: int
    level: int
    set: ArtifactSet


class Constellation(APIModel, Unique):
    """Character constellation."""

    id: int
    icon: str
    pos: int
    name: str
    effect: str
    activated: bool = Aliased("is_actived")

    @property
    def scaling(self) -> bool:
        """Whether the constellation is simply for talent scaling"""
        return "U" in self.icon


class Outfit(APIModel, Unique):
    """Outfit of a character."""

    id: int
    icon: str
    name: str


class Character(PartialCharacter):
    """Character with equipment."""

    weapon: CharacterWeapon
    weapon_type: int


class PropInfo(APIModel):
    """A property such as Crit Rate, HP, HP%."""

    type: int = Aliased("property_type")
    name: str
    icon: typing.Optional[str]
    filter_name: str

    @pydantic.field_validator("name", "filter_name")
    @classmethod
    def __fix_names(cls, value: str) -> str:
        r"""Fix "\xa0" in Crit Damage + Crit Rate names."""
        return value.replace("\xa0", " ")


class PropertyValue(APIModel):
    """A property with a value."""

    base: str
    add: str
    final: str
    info: PropInfo


class DetailCharacterWeapon(CharacterWeapon):
    """Detailed Genshin Weapon with main/sub stats."""

    main_stat: PropertyValue = Aliased("main_property")
    sub_stat: typing.Optional[PropertyValue] = Aliased("sub_property")
    ascension: int = Aliased("promote_level")


class ArtifactProperty(APIModel):
    """Artifact's Property value & roll count."""

    value: str
    times: int
    info: PropInfo


class DetailArtifact(Artifact):
    """Detailed artifact with main/sub stats."""

    main_stat: ArtifactProperty = Aliased("main_property")
    sub_stats: typing.Sequence[ArtifactProperty] = Aliased("sub_property_list")


class SkillAffix(APIModel):
    """Skill affix texts."""

    name: str
    value: str


class CharacterSkill(APIModel):
    """Character's skill."""

    id: int = Aliased("skill_id")
    skill_type: int
    name: str
    level: int

    description: str = Aliased("desc")
    affixes: typing.Sequence[SkillAffix] = Aliased("skill_affix_list")
    icon: str
    is_unlocked: bool = Aliased("is_unlock")


class GenshinDetailCharacter(PartialCharacter):
    """Full Detailed Genshin Character"""

    is_chosen: bool

    # display_image is a different image that is returned by the full character endpoint, but it is not the full gacha art.
    display_image: str = Aliased("image")

    weapon_type: GenshinWeaponType
    weapon: DetailCharacterWeapon

    costumes: typing.Sequence[Outfit]

    artifacts: typing.Sequence[DetailArtifact] = Aliased("relics")
    constellations: typing.Sequence[Constellation]

    skills: typing.Sequence[CharacterSkill]

    selected_properties: typing.Sequence[PropertyValue]
    base_properties: typing.Sequence[PropertyValue]
    extra_properties: typing.Sequence[PropertyValue]
    element_properties: typing.Sequence[PropertyValue]


class GenshinDetailCharacters(APIModel):
    """Genshin character list."""

    characters: typing.Sequence[GenshinDetailCharacter] = Aliased("list")

    property_map: typing.Mapping[str, PropInfo]
    possible_artifact_stats: typing.Mapping[str, typing.Sequence[PropInfo]] = Aliased("relic_property_options")

    artifact_wiki: typing.Mapping[str, str] = Aliased("relic_wiki")
    weapon_wiki: typing.Mapping[str, str]
    avatar_wiki: typing.Mapping[str, str]

    @pydantic.model_validator(mode="before")
    def __fill_prop_info(cls, values: dict[str, typing.Any]) -> typing.Mapping[str, typing.Any]:
        """Fill property info from properety_map."""
        relic_property_options: dict[str, list[int]] = values.get("relic_property_options", {})
        prop_map: dict[str, dict[str, typing.Any]] = values.get("property_map", {})
        characters: list[dict[str, typing.Any]] = values.get("list", [])

        # Map properties to artifacts
        new_relic_prop_options: dict[str, list[dict[str, typing.Any]]] = {}
        for relic_type, properties in relic_property_options.items():
            formatted_properties: list[dict[str, typing.Any]] = [
                prop_map[str(prop)] for prop in properties if str(prop) in prop_map
            ]
            new_relic_prop_options[relic_type] = formatted_properties

        # Override relic_property_options
        values["relic_property_options"] = new_relic_prop_options

        for char in characters:
            # Extract character info from .base
            for key, value in char["base"].items():
                if key == "weapon":  # Ignore .weapon in base as it does not have full info.
                    continue
                char[key] = value

            # Map properties to main/sub stat for weapon.
            main_property = char["weapon"]["main_property"]
            char["weapon"]["main_property"]["info"] = prop_map[str(main_property["property_type"])]
            if sub_property := char["weapon"]["sub_property"]:
                char["weapon"]["sub_property"]["info"] = prop_map[str(sub_property["property_type"])]

            # Map properties to artifacts
            for artifact in char["relics"]:
                main_property = artifact["main_property"]
                artifact["main_property"]["info"] = prop_map[str(main_property["property_type"])]
                for sub_property in artifact["sub_property_list"]:
                    sub_property["info"] = prop_map[str(sub_property["property_type"])]

            # Map character properties
            for prop in (
                char["base_properties"]
                + char["selected_properties"]
                + char["extra_properties"]
                + char["element_properties"]
            ):
                prop["info"] = prop_map[str(prop["property_type"])]

        return values



import datetime

from genshin.models.model import Aliased, APIModel

__all__ = ("WebEvent",)


class WebEvent(APIModel):
    """Hoyolab web event model."""

    id: int
    name: str
    description: str = Aliased("desc")

    start_time: datetime.datetime = Aliased("start")
    end_time: datetime.datetime = Aliased("end")
    create_time: datetime.datetime = Aliased("create_at")

    web_path: str
    app_path: str

    banner: str = Aliased("banner_url")
    status: int

    @property
    def url(self) -> str:
        if "https" in self.web_path:
            return self.web_path
        return f"https://www.hoyolab.com{self.web_path}"



"""Aiohttp webserver used for captcha solving and verification."""

from __future__ import annotations

import asyncio
import typing
import webbrowser

import aiohttp
from aiohttp import web

from genshin import types
from genshin.models.auth.geetest import (
    MMT,
    MMTResult,
    MMTv4,
    MMTv4Result,
    RiskyCheckMMT,
    RiskyCheckMMTResult,
    SessionMMT,
    SessionMMTResult,
    SessionMMTv4,
    SessionMMTv4Result,
)
from genshin.utility import auth as auth_utility

__all__ = ["PAGES", "enter_code", "launch_webapp", "solve_geetest"]

PAGES: typing.Final[dict[typing.Literal["captcha", "enter-code"], str]] = {
    "captcha": """
    <!DOCTYPE html>
    <head>
      <meta name="referrer" content="no-referrer"/>
    </head>
    <html>
      <body></body>
      <script src="./gt/v{gt_version}.js"></script>
      <script>
        const geetestVersion = {gt_version};
        const initGeetest = geetestVersion === 3 ? window.initGeetest : window.initGeetest4;
        fetch("/mmt")
          .then((response) => response.json())
          .then((mmt) => {
            const initParams = geetestVersion === 3 ? {
              gt: mmt.gt,
              challenge: mmt.challenge,
              new_captcha: mmt.new_captcha,
              api_server: "{api_server}",
              https: /^https/i.test(window.location.protocol),
              product: "bind",
              lang: "{lang}",
            } : {
              captchaId: mmt.gt,
              riskType: mmt.risk_type,
              userInfo: mmt.session_id ? JSON.stringify({
                mmt_key: mmt.session_id
              }) : undefined,
              api_server: "{api_server}",
              product: "bind",
              language: "{lang}",
            };
            initGeetest(
              initParams,
              (captcha) => {
                captcha.onReady(() => {
                  geetestVersion == 3 ? captcha.verify() : captcha.showCaptcha();
                });
                captcha.onSuccess(() => {
                  fetch("/send-data", {
                    method: "POST",
                    body: JSON.stringify({
                      ...(mmt.session_id && {session_id: mmt.session_id}),
                      ...(mmt.check_id && {check_id: mmt.check_id}),
                      ...captcha.getValidate()
                    }),
                  }).then(() => window.close());
                  document.body.innerHTML = "You may now close this window.";
                });
              }
            )
          });
      </script>
    </html>
    """,
    "enter-code": """
    <!DOCTYPE html>
    <html>
      <body>
        <input id="code" type="number">
        <button id="verify">Send</button>
      </body>
      <script>
        document.getElementById("verify").onclick = () => {
          fetch("/send-data", {
            method: "POST",
            body: JSON.stringify({
              code: document.getElementById("code").value
            }),
          });
          document.body.innerHTML = "You may now close this window.";
        };
      </script>
    </html>
    """,
}


GT_V3_URL = "https://static.geetest.com/static/js/gt.0.5.0.js"
GT_V4_URL = "https://static.geetest.com/v4/gt4.js"


@typing.overload
async def launch_webapp(
    page: typing.Literal["captcha"],
    *,
    mmt: typing.Union[MMT, MMTv4, SessionMMT, SessionMMTv4, RiskyCheckMMT],
    lang: str = ...,
    api_server: str = ...,
    port: int = ...,
) -> typing.Union[MMTResult, MMTv4Result, SessionMMTResult, SessionMMTv4Result, RiskyCheckMMTResult]: ...
@typing.overload
async def launch_webapp(
    page: typing.Literal["enter-code"],
    *,
    mmt: None = ...,
    lang: None = ...,
    api_server: None = ...,
    port: int = ...,
) -> str: ...
async def launch_webapp(
    page: typing.Literal["captcha", "enter-code"],
    *,
    mmt: typing.Optional[typing.Union[MMT, MMTv4, SessionMMT, SessionMMTv4, RiskyCheckMMT]] = None,
    lang: typing.Optional[str] = None,
    api_server: typing.Optional[str] = None,
    port: int = 5000,
) -> typing.Union[MMTResult, MMTv4Result, SessionMMTResult, SessionMMTv4Result, RiskyCheckMMTResult, str]:
    """Create and run a webapp to solve captcha or enter a verification code."""
    routes = web.RouteTableDef()
    future: asyncio.Future[typing.Any] = asyncio.Future()

    @routes.get("/")
    async def index(request: web.Request) -> web.StreamResponse:
        body = PAGES[page]
        body = body.replace("{gt_version}", "4" if isinstance(mmt, MMTv4) else "3")
        body = body.replace("{api_server}", api_server or "api-na.geetest.com")
        body = body.replace("{lang}", lang or "en")
        return web.Response(body=body, content_type="text/html")

    @routes.get("/gt/{version}.js")
    async def gt(request: web.Request) -> web.StreamResponse:
        version = request.match_info.get("version", "v3")
        gt_url = GT_V4_URL if version == "v4" else GT_V3_URL

        async with aiohttp.ClientSession() as session:
            r = await session.get(gt_url)
            content = await r.read()

        return web.Response(body=content, content_type="text/javascript")

    @routes.get("/mmt")
    async def mmt_endpoint(request: web.Request) -> web.Response:
        return web.json_response(mmt.model_dump() if mmt else {})

    @routes.post("/send-data")
    async def send_data_endpoint(request: web.Request) -> web.Response:
        result = await request.json()
        if "code" in result:
            result = result["code"]
        else:
            if isinstance(mmt, RiskyCheckMMT):
                result = RiskyCheckMMTResult(**result)
            elif isinstance(mmt, SessionMMT):
                result = SessionMMTResult(**result)
            elif isinstance(mmt, SessionMMTv4):
                result = SessionMMTv4Result(**result)
            elif isinstance(mmt, MMT):
                result = MMTResult(**result)
            elif isinstance(mmt, MMTv4):
                result = MMTv4Result(**result)

        future.set_result(result)
        return web.Response(status=204)

    app = web.Application()
    app.add_routes(routes)

    runner = web.AppRunner(app)
    await runner.setup()

    site = web.TCPSite(runner, host="localhost", port=port)
    print(f"Opening http://localhost:{port} in browser...")  # noqa
    webbrowser.open_new_tab(f"http://localhost:{port}")

    await site.start()

    try:
        data = await future
    finally:
        await asyncio.sleep(0.3)
        await runner.shutdown()
        await runner.cleanup()

    return data


@typing.overload
async def solve_geetest(
    mmt: RiskyCheckMMT,
    *,
    lang: types.Lang = ...,
    api_server: str = ...,
    port: int = ...,
) -> RiskyCheckMMTResult: ...
@typing.overload
async def solve_geetest(
    mmt: SessionMMT,
    *,
    lang: types.Lang = ...,
    api_server: str = ...,
    port: int = ...,
) -> SessionMMTResult: ...
@typing.overload
async def solve_geetest(
    mmt: MMT,
    *,
    lang: types.Lang = ...,
    api_server: str = ...,
    port: int = ...,
) -> MMTResult: ...
@typing.overload
async def solve_geetest(
    mmt: SessionMMTv4,
    *,
    lang: types.Lang = ...,
    api_server: str = ...,
    port: int = ...,
) -> SessionMMTv4Result: ...
@typing.overload
async def solve_geetest(
    mmt: MMTv4,
    *,
    lang: types.Lang = ...,
    api_server: str = ...,
    port: int = ...,
) -> MMTv4Result: ...
async def solve_geetest(
    mmt: typing.Union[MMT, MMTv4, SessionMMT, SessionMMTv4, RiskyCheckMMT],
    *,
    lang: types.Lang = "en-us",
    api_server: str = "api-na.geetest.com",
    port: int = 5000,
) -> typing.Union[MMTResult, MMTv4Result, SessionMMTResult, SessionMMTv4Result, RiskyCheckMMTResult]:
    """Start a web server and manually solve geetest captcha."""
    geetest_lang = auth_utility.lang_to_geetest_lang(lang)
    return await launch_webapp(
        "captcha",
        mmt=mmt,
        lang=geetest_lang,
        api_server=api_server,
        port=port,
    )


async def enter_code(*, port: int = 5000) -> str:
    """Get email or phone number verification code from user."""
    return await launch_webapp("enter-code", port=port)



import typing
from enum import Enum

from genshin.models.model import Aliased, APIModel, TZDateTime

__all__ = (
    "IncomeData",
    "PolychromeIncome",
    "PolychromeIncomeType",
    "ZZZCurrencyType",
    "ZZZDiary",
    "ZZZDiaryDetail",
    "ZZZDiaryDetailItem",
    "ZZZDiaryPlayerInfo",
    "ZZZIncomeCurrency",
)


class PolychromeIncomeType(Enum):
    """Polychrome income type enum."""

    DAILY = "daily_activity_rewards"
    DEVELOPMENT = "growth_rewards"
    SHIYU_DEFENSE = "shiyu_rewards"
    MAIL = "mail_rewards"
    HOLLOW_ZERO = "hollow_rewards"
    EVENT = "event_rewards"
    OTHER = "other_rewards"


class PolychromeIncome(APIModel):
    """Polychrome income model."""

    source: PolychromeIncomeType = Aliased("action")
    num: int
    percent: int


class ZZZCurrencyType(Enum):
    """Currency type enum."""

    POLYCHROME = "PolychromesData"
    MASTER_TAPE = "MatserTapeData"
    BOOPONS = "BooponsData"


class ZZZIncomeCurrency(APIModel):
    """Currency model."""

    type: ZZZCurrencyType = Aliased("data_type")
    num: int = Aliased("count")
    name: str = Aliased("data_name")


class IncomeData(APIModel):
    """Income model."""

    currencies: typing.Sequence[ZZZIncomeCurrency] = Aliased("list")
    polychrome_incomes: typing.Sequence[PolychromeIncome] = Aliased("income_components")


class ZZZDiaryPlayerInfo(APIModel):
    """Player info model."""

    nickname: str
    avatar_url: str = Aliased("avatar")


class ZZZDiary(APIModel):
    """ZZZ monthly earnings model."""

    uid: int
    region: str
    current_month: str
    data_month: str

    income: IncomeData = Aliased("month_data")
    month_options: typing.Sequence[str] = Aliased("optional_month")
    player: ZZZDiaryPlayerInfo = Aliased("role_info")


class ZZZDiaryDetailItem(APIModel):
    """ZZZ monthly earning detail item model."""

    id: str
    source: PolychromeIncomeType = Aliased("action")
    time: TZDateTime
    num: int


class ZZZDiaryDetail(APIModel):
    """ZZZ monthly earning detail model."""

    uid: int
    region: str
    data_month: str
    current_page: int
    total: int
    items: typing.Sequence[ZZZDiaryDetailItem] = Aliased("list")

    currency_name: str = Aliased("data_name")
    currency_type: ZZZCurrencyType = Aliased("data_type")
